import torch.nn as nn
import math
from tqdm import tqdm
import numpy as np
from dalle_pytorch import DiscreteVAE

from vilmedic.models.utils import get_n_params


def evaluation(models, config, dl, **kwargs):
    losses = []
    model = models[0]
    pbar = tqdm(dl, total=len(dl))
    for batch in pbar:
        out = model(**batch)
        out['loss'] = out['loss'].mean()
        losses.append(out['loss'].cpu().data.numpy())
    return {'loss': np.ndarray.mean(np.array(losses))}


class VAE(nn.Module):
    def __init__(self,
                 image_size=256,
                 num_layers=3,
                 num_tokens=8192,
                 codebook_dim=512,
                 hidden_dim=64,
                 num_resnet_blocks=1,
                 temperature=0.9,
                 straight_through=False,
                 **kwargs):
        super(VAE, self).__init__()

        self.vae = DiscreteVAE(
            image_size=image_size,
            num_layers=num_layers,
            num_tokens=num_tokens,
            codebook_dim=codebook_dim,
            hidden_dim=hidden_dim,
            num_resnet_blocks=num_resnet_blocks,
            temperature=temperature,
            straight_through=straight_through
        )

        self.global_step = 0
        self.temp = temperature

        self.eval_func = evaluation

    def forward(self, images, **kwargs):
        loss, output = self.vae(images.cuda(),
                                return_loss=True,
                                return_recons=True,
                                temp=self.temp
                                )
        self.global_step += 1
        if self.global_step % 100 == 0:
            self.temp = max(self.temp * math.exp(-1e-6 * self.global_step), 0.5)

        return {'loss': loss, 'output': output}

    def __repr__(self):
        s = "DiscreteVAE()" + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import torch.nn as nn
import torchvision


class vgg16HGap(nn.Module):
    def __init__(self, pretrained=True):
        super(vgg16HGap, self).__init__()
        self.vgg = torchvision.models.vgg16(pretrained=pretrained).features
        self.model_list = list(self.vgg.children())
        self.conv1 = nn.Sequential(*self.model_list[0:4])
        self.conv2 = nn.Sequential(*self.model_list[4:9])
        self.conv3 = nn.Sequential(*self.model_list[9:16])
        self.conv4 = nn.Sequential(*self.model_list[16:23])
        self.conv5 = nn.Sequential(*self.model_list[23:30])
        self.conv6 = nn.Sequential(self.model_list[30])
        self.pool1 = nn.AdaptiveAvgPool2d(1)
        self.pool2 = nn.AdaptiveAvgPool2d(1)
        self.pool3 = nn.AdaptiveAvgPool2d(1)
        self.pool4 = nn.AdaptiveAvgPool2d(1)
        self.pool5 = nn.AdaptiveAvgPool2d(1)
        self.pool6 = nn.AdaptiveAvgPool2d(1)

        # self.lsoftmax_linear = LSoftmaxLinear(
        #     input_features=1984, output_features=num_classes, margin=margin, device=rank)

        # self.asoftmax = AngleLinear(1984, num_classes, m=margin)

    def forward(self, x, targeta=None, targetb=None):
        y1 = self.conv1(x)
        p1 = self.pool1(y1)

        y2 = self.conv2(y1)
        p2 = self.pool2(y2)

        y3 = self.conv3(y2)
        p3 = self.pool3(y3)

        y4 = self.conv4(y3)
        p4 = self.pool4(y4)

        y5 = self.conv5(y4)
        p5 = self.pool5(y5)

        y6 = self.conv6(y5)
        p6 = self.pool6(y6)

        f = torch.cat([p1, p2, p3, p4, p5, p6], 1).squeeze(3).squeeze(2)

        return f

        # if self.training:
        #     outa = self.lsoftmax_linear(input=f, target=targeta)
        #     outb = self.lsoftmax_linear(input=f, target=targetb)
        #     return outa, outb
        # else:
        #     return self.lsoftmax_linear(input=f)

        # return self.asoftmax(f)


class vgg19HGap(nn.Module):
    def __init__(self, pretrained=True):
        super(vgg19HGap, self).__init__()
        self.vgg = torchvision.models.vgg19(pretrained=pretrained).features
        self.model_list = list(self.vgg.children())
        self.conv1 = nn.Sequential(*self.model_list[0:4])
        self.conv2 = nn.Sequential(*self.model_list[4:9])
        self.conv3 = nn.Sequential(*self.model_list[9:18])
        self.conv4 = nn.Sequential(*self.model_list[18:27])
        self.conv5 = nn.Sequential(*self.model_list[27:36])
        self.conv6 = nn.Sequential(self.model_list[36])
        self.pool1 = nn.AvgPool2d(224)
        self.pool2 = nn.AvgPool2d(112)
        self.pool3 = nn.AvgPool2d(56)
        self.pool4 = nn.AvgPool2d(28)
        self.pool5 = nn.AvgPool2d(14)
        self.pool6 = nn.AvgPool2d(7)


    def forward(self, x):
        y1 = self.conv1(x)
        p1 = self.pool1(y1)

        y2 = self.conv2(y1)
        p2 = self.pool2(y2)

        y3 = self.conv3(y2)
        p3 = self.pool3(y3)

        y4 = self.conv4(y3)
        p4 = self.pool4(y4)

        y5 = self.conv5(y4)
        p5 = self.pool5(y5)

        y6 = self.conv6(y5)
        p6 = self.pool6(y6)

        out = torch.cat([p1, p2, p3, p4, p5, p6], 1).squeeze(3)
        out = out.squeeze(2)
        return out



from .Seq2Seq import Seq2Seq
from .ImSeq2Seq import ImSeq2Seq
from .ImLabel import ImLabel
from .ImSeqLabel import ImSeqLabel
from .ImSeqAny import ImSeqAny
from .ImSeq import ImSeq
from .base.ImageDataset import ImageDataset
from .base.TextDataset import TextDataset
from .base.LabelDataset import LabelDataset
from .base.papers.IndexDataset import IndexDataset
from .base.papers.stanford_eeg.TUHDataset import TUHDataset



from .selfsup.VICREGLoss import VICREGLoss
from .selfsup.ConVIRTLoss import ConVIRTLoss
from .selfsup.InfoNCELoss import InfoNCELoss
from .selfsup.GLoRIALoss import GLoRIALoss, gloria_attention_fn, cosine_similarity
from .mvqa.LabelSmoothingCrossEntropyLoss import LabelSmoothingCrossEntropy
from torch.nn.modules.loss import *



from ... import *


class ActualIndexDataset:
    """
    Dataset wrapper that returns the index of the samples
    """
    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {
                **super(ActualIndexDataset, self).get_collate_fn()(batch),
                "index": [s['index'] for s in batch]
            }
            return collated

        return collate_fn

    def __getitem__(self, index):
        return {**super(ActualIndexDataset, self).__getitem__(index), 'index': index}

    def __repr__(self):
        return "IndexDataset with original dataset being: \n" + \
               super(ActualIndexDataset, self).__repr__()


def IndexDataset(dataset, **kwargs):
    return type('ActualIndexDataset', (ActualIndexDataset, eval(dataset), object), {})(**kwargs)



# https://github.com/PyTorchLightning/lightning-bolts/blob/master/pl_bolts/models/self_supervised/swav/transforms.py
from typing import Tuple

import numpy as np

from pl_bolts.utils import _OPENCV_AVAILABLE, _TORCHVISION_AVAILABLE
from pl_bolts.utils.warnings import warn_missing_pkg

if _TORCHVISION_AVAILABLE:
    from torchvision import transforms
else:  # pragma: no cover
    warn_missing_pkg("torchvision")

if _OPENCV_AVAILABLE:
    import cv2
else:  # pragma: no cover
    warn_missing_pkg("cv2", pypi_name="opencv-python")


class SwAVTrainDataTransform:
    def __init__(
            self,
            normalize=None,
            size_crops: Tuple[int] = (96, 36),
            nmb_crops: Tuple[int] = (2, 4),
            min_scale_crops: Tuple[float] = (0.33, 0.10),
            max_scale_crops: Tuple[float] = (1, 0.33),
            gaussian_blur: bool = True,
            jitter_strength: float = 1.0,
    ):
        self.jitter_strength = jitter_strength
        self.gaussian_blur = gaussian_blur

        assert len(size_crops) == len(nmb_crops)
        assert len(min_scale_crops) == len(nmb_crops)
        assert len(max_scale_crops) == len(nmb_crops)

        self.size_crops = size_crops
        self.nmb_crops = nmb_crops
        self.min_scale_crops = min_scale_crops
        self.max_scale_crops = max_scale_crops

        self.color_jitter = transforms.ColorJitter(
            0.8 * self.jitter_strength,
            0.8 * self.jitter_strength,
            0.8 * self.jitter_strength,
            0.2 * self.jitter_strength,
        )

        transform = []
        color_transform = [transforms.RandomApply([self.color_jitter], p=0.8), transforms.RandomGrayscale(p=0.2)]

        if self.gaussian_blur:
            kernel_size = int(0.1 * self.size_crops[0])
            if kernel_size % 2 == 0:
                kernel_size += 1

            color_transform.append(GaussianBlur(kernel_size=kernel_size, p=0.5))

        self.color_transform = transforms.Compose(color_transform)

        if normalize is None:
            self.final_transform = transforms.ToTensor()
        else:
            self.final_transform = transforms.Compose([transforms.ToTensor(), normalize])

        for i in range(len(self.size_crops)):
            random_resized_crop = transforms.RandomResizedCrop(
                self.size_crops[i],
                scale=(self.min_scale_crops[i], self.max_scale_crops[i]),
            )

            transform.extend(
                [
                    transforms.Compose(
                        [
                            random_resized_crop,
                            transforms.RandomHorizontalFlip(p=0.5),
                            self.color_transform,
                            self.final_transform,
                        ]
                    )
                ]
                * self.nmb_crops[i]
            )

        self.transform = transform

        # add online train transform of the size of global view
        online_train_transform = transforms.Compose(
            [transforms.RandomResizedCrop(self.size_crops[0]), transforms.RandomHorizontalFlip(), self.final_transform]
        )

        self.transform.append(online_train_transform)

    def __call__(self, sample):
        multi_crops = list(map(lambda transform: transform(sample), self.transform))

        return multi_crops


class SwAVEvalDataTransform(SwAVTrainDataTransform):
    def __init__(
            self,
            normalize=None,
            size_crops: Tuple[int] = (96, 36),
            nmb_crops: Tuple[int] = (2, 4),
            min_scale_crops: Tuple[float] = (0.33, 0.10),
            max_scale_crops: Tuple[float] = (1, 0.33),
            gaussian_blur: bool = True,
            jitter_strength: float = 1.0,
    ):
        super().__init__(
            normalize=normalize,
            size_crops=size_crops,
            nmb_crops=nmb_crops,
            min_scale_crops=min_scale_crops,
            max_scale_crops=max_scale_crops,
            gaussian_blur=gaussian_blur,
            jitter_strength=jitter_strength,
        )

        input_height = self.size_crops[0]  # get global view crop
        test_transform = transforms.Compose(
            [
                transforms.Resize(int(input_height + 0.1 * input_height)),
                transforms.CenterCrop(input_height),
                self.final_transform,
            ]
        )

        # replace last transform to eval transform in self.transform list
        self.transform[-1] = test_transform


class SwAVFinetuneTransform:
    def __init__(
            self, input_height: int = 224, jitter_strength: float = 1.0, normalize=None, eval_transform: bool = False
    ) -> None:

        self.jitter_strength = jitter_strength
        self.input_height = input_height
        self.normalize = normalize

        self.color_jitter = transforms.ColorJitter(
            0.8 * self.jitter_strength,
            0.8 * self.jitter_strength,
            0.8 * self.jitter_strength,
            0.2 * self.jitter_strength,
        )
        if not eval_transform:
            data_transforms = [
                transforms.RandomResizedCrop(size=self.input_height),
                transforms.RandomHorizontalFlip(p=0.5),
                transforms.RandomApply([self.color_jitter], p=0.8),
                transforms.RandomGrayscale(p=0.2),
            ]
        else:
            data_transforms = [
                transforms.Resize(int(self.input_height + 0.1 * self.input_height)),
                transforms.CenterCrop(self.input_height),
            ]

        if normalize is None:
            final_transform = transforms.ToTensor()
        else:
            final_transform = transforms.Compose([transforms.ToTensor(), normalize])

        data_transforms.append(final_transform)
        self.transform = transforms.Compose(data_transforms)

    def __call__(self, sample):
        return self.transform(sample)


class GaussianBlur:
    # Implements Gaussian blur as described in the SimCLR paper
    def __init__(self, kernel_size, p=0.5, min=0.1, max=2.0):
        self.min = min
        self.max = max

        # kernel size is set to be 10% of the image height/width
        self.kernel_size = kernel_size
        self.p = p

    def __call__(self, sample):
        sample = np.array(sample)

        # blur the image with a 50% chance
        prob = np.random.random_sample()

        if prob < self.p:
            sigma = (self.max - self.min) * np.random.random_sample() + self.min
            sample = cv2.GaussianBlur(sample, (self.kernel_size, self.kernel_size), sigma)

        return sample



import torch
import copy
import os
import operator
import re
import json
import inspect

import numpy as np
import torch.nn as nn

from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torch.utils.data.sampler import BatchSampler, SequentialSampler, RandomSampler

from vilmedic.models import *
from vilmedic.datasets import *

import torch_optimizer
from torch.optim import *
from torch_optimizer import *
from torch.optim.lr_scheduler import *
from vilmedic.blocks.schedulers import LinearWarmupCosineAnnealingLR


def get_eval_func(models):
    dummy = models[0]
    if isinstance(dummy, nn.DataParallel):
        dummy = dummy.module
    assert hasattr(dummy, "eval_func")
    return dummy.eval_func


def create_optimizer(config, logger, params, state_dict=None):
    assert 'lr' in config.optim_params
    if hasattr(torch.optim, config.optimizer):
        optim = getattr(torch.optim, config.optimizer)
    elif hasattr(torch_optimizer, config.optimizer):
        optim = getattr(torch_optimizer, config.optimizer)
    else:
        raise NotImplementedError(config.optimizer)

    optimizer = optim(params, **config.optim_params)
    logger.settings('Optimizer {} created'.format(type(optimizer).__name__))

    if state_dict is not None and "optimizer" in state_dict:
        optimizer.load_state_dict(state_dict["optimizer"])
        logger.info('Optimizer state loaded')
    else:
        logger.info(optimizer)
    return optimizer


def create_model(config, dl, logger, from_training=True, state_dict=None):
    # Create model, give him dataloader also
    config = copy.deepcopy(config.model)
    model = eval(config.pop('proto'))(**config, dl=dl, logger=logger, from_training=from_training)
    logger.settings('Model {} created'.format(type(model).__name__))

    if state_dict is not None and "model" in state_dict:
        params = {k.replace('module.', ''): v for k, v in state_dict["model"].items()}
        model.load_state_dict(params)
        logger.info('Model state loaded')
    else:
        logger.info(model)

    model = model.cuda()

    if torch.cuda.device_count() > 1:
        logger.info("Using {} GPUs!".format(torch.cuda.device_count()))
        model = torch.nn.DataParallel(model)
    return model


def create_data_loader(config, split, logger, called_by_validator=False, called_by_ensemblor=False):
    dataset_config = copy.deepcopy(config.dataset)
    # Its important the dataset receive info if call from ensemblor (test time):
    # split can be train with validation transformation
    dataset = eval(dataset_config.proto)(split=split,
                                         ckpt_dir=config.ckpt_dir,
                                         called_by_ensemblor=called_by_ensemblor,
                                         **dataset_config)

    if hasattr(dataset, 'get_collate_fn'):
        collate_fn = dataset.get_collate_fn()
    else:
        collate_fn = default_collate

    # RandomSampler for train split, during training only
    if split == 'train' and not called_by_validator:
        sampler = BatchSampler(
            RandomSampler(dataset),
            batch_size=config.batch_size,
            drop_last=config.drop_last or False)
        logger.info('Using ' + type(sampler.sampler).__name__)

    else:
        sampler = BatchSampler(
            SequentialSampler(dataset),
            batch_size=config.batch_size,
            drop_last=False)

    # Print dataset for training and ensemblor
    if not called_by_validator or called_by_ensemblor:
        logger.settings('DataLoader')
        logger.info(dataset)

    return DataLoader(dataset,
                      num_workers=dataset_config.num_workers or 4,
                      collate_fn=collate_fn,
                      batch_sampler=sampler,
                      pin_memory=True)


def create_scaler(config, logger, state_dict=None):
    scaler = torch.cuda.amp.GradScaler(enabled=(config.use_amp or False))
    logger.settings('Using scaler : {}'.format(scaler.is_enabled()))
    if state_dict is not None and "scaler" in state_dict:
        scaler.load_state_dict(state_dict["scaler"])
        logger.info('Scaler state loaded')
    return scaler


def create_training_scheduler(config, optimizer, logger, state_dict=None):
    config = copy.deepcopy(config)
    training_scheduler = TrainingScheduler(lr_decay_func=config.lr_decay,
                                           optimizer=optimizer,
                                           early_stop_metric=config.early_stop_metric,
                                           early_stop_limit=config.early_stop,
                                           lr_decay_params=config.lr_decay_params)
    logger.settings('Training scheduler created')
    if state_dict is not None and "training_scheduler" in state_dict:
        training_scheduler.load_state_dict(state_dict["training_scheduler"])
        logger.info('Training scheduler state loaded')
    else:
        logger.info(training_scheduler)
    return training_scheduler


class CheckpointSaver(object):
    def __init__(self, ckpt_dir, logger, seed, ckpt=None):
        self.ckpt_dir = ckpt_dir
        self.seed = seed
        self.logger = logger
        self.current_tag = None
        self.current_epoch = None

        if ckpt is not None:
            self.current_tag, self.current_epoch = self.extract_tag_and_step(ckpt)
            self.logger.settings(
                'Resuming checkpoint after epoch {} with tag {}.'.format(self.current_epoch + 1, self.current_tag))

    def save(self, state_dict, tag, current_epoch):
        if self.current_tag is not None:
            old_ckpt = os.path.join(self.ckpt_dir,
                                    '{}_{}_{}.pth'.format(self.current_tag, self.current_epoch, self.seed))
            assert os.path.exists(old_ckpt), old_ckpt
            os.remove(old_ckpt)

        tag = np.round(tag, 6)
        path = os.path.join(self.ckpt_dir, '{}_{}_{}.pth'.format(tag, current_epoch, self.seed))
        torch.save(state_dict, path)
        self.logger.info('{} saved.'.format(path))

        self.current_tag = tag
        self.current_epoch = current_epoch

    def extract_tag_and_step(self, ckpt):
        groups = re.match('.*/(.*?)_(.*?)_(.*?).pth', ckpt)
        return float(groups.group(1)), int(groups.group(2))


class TrainingScheduler(object):
    iter_step_scheduler = {"CyclicLR", "OneCycleLR", "CosineAnnealingWarmRestarts"}
    epoch_step_scheduler = {"LambdaLR", "MultiplicativeLR", "StepLR", "MultiStepLR", "ConstantLR", "LinearLR",
                            "ExponentialLR", "ChainedScheduler", "SequentialLR", "CosineAnnealingLR",
                            "LinearWarmupCosineAnnealingLR"}
    val_step_scheduler = {"ReduceLROnPlateau"}

    def __init__(self, lr_decay_func, optimizer, early_stop_metric, early_stop_limit, lr_decay_params):
        super().__init__()

        self.epoch = 0
        self.early_stop = 0
        self.early_stop_limit = early_stop_limit
        self.metric_comp_func = operator.gt
        self.mode = 'max'
        self.current_best_metric = -float('inf')
        self.lr_decay_params = lr_decay_params
        self.early_stop_metric = early_stop_metric

        # 4info: You can decay_on_training_loss and have a early_stop_metric different than training loss
        self.decay_on_training_loss = self.lr_decay_params.decay_on_training_loss or False

        if early_stop_metric in ['validation_loss', 'training_loss']:
            self.metric_comp_func = operator.lt
            self.mode = 'min'
            self.current_best_metric = float('inf')

        self.scheduler_name = lr_decay_func
        if self.scheduler_name == 'ReduceLROnPlateau':
            self.lr_decay_params["mode"] = self.mode

        def remove_unused_args(func, **kwargs):
            sig = [param.name for param in inspect.signature(func).parameters.values()]
            return {k: v for k, v in kwargs.items() if k in sig}

        self.lr_decay_params = remove_unused_args(eval(lr_decay_func), **self.lr_decay_params)
        self.scheduler = eval(lr_decay_func)(optimizer, **self.lr_decay_params)

    def iteration_step(self):
        if self.scheduler_name in TrainingScheduler.iter_step_scheduler:
            self.scheduler.step()

    def epoch_step(self):
        self.epoch = self.epoch + 1
        if self.scheduler_name in TrainingScheduler.epoch_step_scheduler:
            self.scheduler.step()

    def eval_step(self, decay_metric=None, early_stop_score=None):
        ret = {
            "done_training": False,
            "save_state": False,
        }

        # LR scheduler
        if decay_metric is not None:
            if self.scheduler_name in TrainingScheduler.val_step_scheduler:
                self.scheduler.step(decay_metric)

        # Early stop
        if early_stop_score is not None:
            if self.metric_comp_func(early_stop_score, self.current_best_metric):
                self.current_best_metric = early_stop_score
                self.early_stop = 0
                ret["save_state"] = True
            else:
                self.early_stop += 1
                if self.early_stop == self.early_stop_limit:
                    ret["done_training"] = True
        return ret

    def __repr__(self):
        s = "TrainingScheduler (\n"
        s += self.scheduler_name + "\n"
        s += str(json.dumps(dict(self.lr_decay_params), indent=4, sort_keys=True)) + '\n'
        s += 'Early stopping' + "\n"
        s += '    {0}: {1}\n'.format("early_stop_limit", self.early_stop_limit)
        s += '    {0}: {1}\n'.format("metric_comp_func", self.metric_comp_func)
        s += '    {0}: {1}\n'.format("mode", self.mode)
        s += '    {0}: {1}\n'.format("current_best_metric", self.current_best_metric)
        s += '    {0}: {1}\n'.format("decay_on_training_loss", self.decay_on_training_loss)
        s += ')'
        return s

    def state_dict(self):
        training_sched = {key: value for key, value in self.__dict__.items() if key != 'scheduler'}
        training_sched["scheduler"] = self.scheduler.state_dict()
        return training_sched

    def load_state_dict(self, state_dict):
        if "scheduler" in state_dict:  # Retro compatible with older checkpoint version
            scheduler = state_dict.pop("scheduler")
            self.__dict__.update(state_dict)
            self.scheduler.load_state_dict(scheduler)



# Filename: ciderD.py
#
# Description: Describes the class to compute the CIDEr-D (Consensus-Based Image Description Evaluation) Metric
#               by Vedantam, Zitnick, and Parikh (http://arxiv.org/abs/1411.5726)
#
# Creation Date: Sun Feb  8 14:16:54 2015
#
# Authors: Ramakrishna Vedantam <vrama91@vt.edu> and Tsung-Yi Lin <tl483@cornell.edu>
from .ciderD_RL_scorer import CiderScorer
from vilmedic.datasets.base.papers.report_preprocessing import *


class CiderDRL:
    """
    Main Class to compute the CIDEr metric

    """

    def __init__(self, n=4, sigma=6.0, df="corpus", processing=None, **kwargs):
        # set cider to sum over 1 to 4-grams
        super().__init__()
        self._n = n
        # set the standard deviation parameter for gaussian penalty
        self._sigma = sigma
        # set which where to compute document frequencies from
        refs = [eval(processing or 'lambda x: x')(ref.strip()) for ref in open(df).readlines()]
        scorer = CiderScorer(refs=refs)
        scorer.compute_doc_freq()
        self._df = scorer.document_frequency

    def __call__(self, gts, res):
        """
        Main function to compute CIDEr score
        :param  hypo_for_image (dict) : dictionary with key <image> and value <tokenized hypothesis / candidate sentence>
                ref_for_image (dict)  : dictionary with key <image> and value <tokenized reference sentence>
        :return: cider (float) : computed CIDEr score for the corpus
        """

        cider_scorer = CiderScorer(n=self._n, sigma=self._sigma, df=self._df)
        res = {i: [v] for i, v in enumerate(res)}
        gts = {i: [v] for i, v in enumerate(gts)}
        for id in sorted(gts.keys()):
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert (type(hypo) is list)
            assert (len(hypo) == 1)
            assert (type(ref) is list)
            assert (len(ref) > 0)
            cider_scorer += (hypo[0], ref)

        (score, scores) = cider_scorer.compute_score()

        return score, scores

    def method(self):
        return "CIDEr-D"


if __name__ == '__main__':
    import pickle

    x = CiderDRL(df="../../../../../data/RRG/mimic-cxr/impression/train.impression.tok")
    for i in range(200000):
        x(
            gts=[
                'no evidence of consolidation to suggest pneumonia is seen. there  is some retrocardiac atelectasis. a small left pleural effusion may be  present. no pneumothorax is seen. no pulmonary edema. a right granuloma is  unchanged. the heart is mildly enlarged, unchanged. there is tortuosity of  the aorta.',
                'there are moderate bilateral pleural effusions with overlying atelectasis,  underlying consolidation not excluded. mild prominence of the interstitial  markings suggests mild pulmonary edema. the cardiac silhouette is mildly  enlarged. the mediastinal contours are unremarkable. there is no evidence of  pneumothorax.'
            ],
            res=[
                'heart size is moderately enlarged. the mediastinal and hilar contours are unchanged. there is no pulmonary edema. small left pleural effusion is present. patchy opacities in the lung bases likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.',
                'heart size is mildly enlarged. the mediastinal and hilar contours are normal. there is mild pulmonary edema. moderate bilateral pleural effusions are present, left greater than right. bibasilar airspace opacities likely reflect atelectasis. no pneumothorax is seen. there are no acute osseous abnormalities.'
            ])



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params
import numpy as np

from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation
from vilmedic.blocks.vision.eeg_modeling.dense_inception import DenseInception

from einops import rearrange


class RRG_EEG(nn.Module):

    def __init__(self, decoder, dl, pretrained=False, freeze=False,
                 **kwargs):
        super().__init__()
        # Decoder
        self.dec = DecoderModel(decoder)

        # Encoder
        eeg_enc = DenseInception(num_classes=768, data_shape=(12000, 19))
        if pretrained:
            pretrained_pth = "/home/ksaab/Documents/eeg_fully_supervised/results/lpch_mini2/scale_1/cv_3/best.pth.tar"
            state_dict = torch.load(pretrained_pth)["state_dict"]
            og_state_dict = eeg_enc.state_dict()
            # clean up state dict keys
            state_dict_ = {}
            for key in state_dict:
                if "fc2" not in key:
                    state_dict_[key.split("dense_inception.")[-1]] = state_dict[key]
                else:
                    state_dict_[key.split("dense_inception.")[-1]] = og_state_dict[
                        key.split("dense_inception.")[-1]
                    ]
            eeg_enc.load_state_dict(state_dict_)
        # eeg_enc.fc2 = torch.nn.Identity()
        if freeze:
            for param in eeg_enc.parameters():
                param.requires_grad = False

        # visual_projection = nn.Linear(360, self.dec.decoder.config.hidden_size)
        # self.enc = nn.Sequential(eeg_enc, visual_projection)
        self.enc = eeg_enc

        # Evaluation
        self.eval_func = evaluation
        self.tokenizer = dl.dataset.tokenizer

    def forward(self, input_ids, attention_mask, images, images_mask=None, encoder_outputs=None,
                encoder_attention_mask=None, epoch=None, iteration=None, **kwargs):

        if encoder_outputs is None:
            encoder_outputs, encoder_attention_mask = self.encode(images, images_mask, **kwargs)

        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()

        out = self.dec(input_ids=input_ids,
                       attention_mask=attention_mask,
                       encoder_outputs=encoder_outputs,
                       encoder_attention_mask=encoder_attention_mask,
                       **kwargs)
        # print(self.tokenizer.decode(torch.argmax(out["logits"][0], dim=-1), skip_special_tokens=True,
        #                             clean_up_tokenization_spaces=False))

        return out

    # Necessary for generation
    def encode(self, images, images_mask=None, **kwargs):
        eeg = images.cuda()
        images_mask = images_mask.cuda()

        # eeg.shape torch.Size([bs, 30, 60, 19, 200])
        eeg = rearrange(eeg, 'd0 d1 d2 d3 d4 -> (d0 d1) (d2 d4) d3')
        eeg = eeg.cuda()
        print(eeg.shape)
        feature = self.enc(eeg)
        print(feature.shape)
        stop
        return feature, images_mask

    def __repr__(self):
        s = "model: RRG_EEG\n"
        s += "(enc):" + str(self.enc) + '\n'
        s += "(dec):" + str(self.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/02-ppo.ipynb (unless otherwise specified).

__all__ = ['AdaptiveKLController', 'FixedKLController', 'PPOTrainer', 'stack_dicts', 'stats_to_np']

# Cell
import numpy as np
import torch.nn.functional as F
from torch.optim import Adam
import torch
import collections
import time
import random

from .core import (logprobs_from_logits,
                   whiten,
                   clip_by_value,
                   entropy_from_logits,
                   flatten_dict,
                   average_torch_dicts,
                   stats_to_np,
                   stack_dicts,
                   add_suffix)
from collections import defaultdict
from torch_optimizer import *

# Cell

def first_zero(x, axis=0):
    z = (x == 0.)
    return ((z.cumsum(axis) == 1) & z).max(axis)


class AdaptiveKLController:
    """
    Adaptive KL controller described in the paper:
    https://arxiv.org/pdf/1909.08593.pdf
    """

    def __init__(self, init_kl_coef, target, horizon):
        self.value = init_kl_coef
        self.target = target
        self.horizon = horizon

    def update(self, current, n_steps):
        target = self.target
        proportional_error = np.clip(current / target - 1, -0.2, 0.2)
        mult = 1 + proportional_error * n_steps / self.horizon
        self.value *= mult


# Cell

class FixedKLController:
    """Fixed KL controller."""

    def __init__(self, kl_coef):
        self.value = kl_coef

    def update(self, current, n_steps):
        pass


# Cell

class PPOTrainer:
    """
    The PPO_trainer uses Proximal Policy Optimization to optimise language models.
    """

    default_params = {
        "lr": 5e-6,
        "adap_kl_ctrl": True,
        "init_kl_coef": 0.2,
        "target": 6,
        "horizon": 10000,
        "gamma": 1,
        "lam": 0.95,
        "cliprange": .2,
        "cliprange_value": .2,
        "vf_coef": .1,
        "forward_batch_size": 2,
        "ppo_epochs": 4,
        "top_k": 0,
        "top_p": 1.0
    }

    def __init__(self, encoder, model, ref_model, v_head, **ppo_params):
        """
        Initialize PPOTrainer.

        Args:
            model (torch.model): Hugging Face transformer GPT2 model with value head
            ref_model (torch.model): Hugging Face transformer GPT2 refrence model used for KL penalty
            ppo_params (dict or None): PPO parameters for training. Can include following keys:
                'lr' (float): Adam learning rate, default: 1.41e-5
                'batch_size' (int): Number of samples per optimisation step, default: 256
                'forward_batch_size' (int): Number of samples forward passed through model at a time, default: 16
                'ppo_epochs' (int): Number of optimisation epochs per batch of samples, default: 4
                'gamma' (float)): Gamma parameter for advantage calculation, default: 1.
                'lam' (float): Lambda parameter for advantage calcualation, default: 0.95
                'cliprange_value' (float): Range for clipping values in loss calculation, default: 0.2
                'cliprange' (float): Range for clipping in PPO policy gradient loss, default: 0.2
                'vf_coef' (float): Scaling factor for value loss, default: 0.1
                'adap_kl_ctrl' (bool): Use adaptive KL control, otherwise linear, default: True
                'init_kl_coef' (float): Initial KL penalty coefficient (used for adaptive and linear control), default: 0.2
                'target' (float): Target KL value for adaptive KL control, default: 6.0
                'horizon' (float): Horizon for adaptive KL control, default: 10000

        """
        self.ppo_params = self.default_params
        self.ppo_params.update(ppo_params)

        self.ref_model = ref_model
        self.model = model
        self.encoder = encoder
        self.v_head = v_head
        self.optimizer = RAdam(model.parameters(), lr=self.ppo_params['lr'])

        self.kl_ctl = AdaptiveKLController(self.ppo_params['init_kl_coef'],
                                           self.ppo_params['target'],
                                           self.ppo_params['horizon'])

        self.stats = defaultdict(list)

    def step(self, query, response, scores):
        """
        Run a PPO optimisation step.

        args:
            query (torch.tensor): tensor containing the encoded queries, shape [batch_size, query_length]
            response (torch.tensor): tensor containing the encoded responses, shape [batch_size, response_length]
            scores (torch.tensor): tensor containing the scores, shape [batch_size]

        returns:
            train_stats (dict): a summary of the training statistics
        """

        bs = self.ppo_params['batch_size']
        timing = dict()
        t0 = time.time()

        gen_len = response.shape[1]
        model_input = torch.cat((query, response), axis=1)

        t = time.time()
        logprobs, ref_logprobs, values = self.batched_forward_pass(model_input, gen_len)
        timing['time/ppo/forward_pass'] = time.time() - t

        t = time.time()
        rewards, non_score_reward, kl_coef = self.compute_rewards(scores, logprobs, ref_logprobs)
        timing['time/ppo/compute_rewards'] = time.time() - t

        t = time.time()
        all_stats = []
        idxs = list(range(bs))
        for _ in range(self.ppo_params['ppo_epochs']):
            random.shuffle(idxs)
            for i in range(bs):
                idx = idxs[i]
                train_stats = self.train_minibatch(logprobs[idx:idx + 1], values[idx:idx + 1],
                                                   rewards[idx:idx + 1], query[idx:idx + 1],
                                                   response[idx:idx + 1], model_input[idx:idx + 1])
                all_stats.append(train_stats)
        timing['time/ppo/optimize_step'] = time.time() - t

        t = time.time()
        train_stats = stack_dicts(all_stats)

        # reshape advantages/ratios such that they are not averaged.
        train_stats['policy/advantages'] = torch.flatten(train_stats['policy/advantages']).unsqueeze(0)
        train_stats['policy/ratio'] = torch.flatten(train_stats['policy/ratio']).unsqueeze(0)

        stats = self.record_step_stats(scores=scores, logprobs=logprobs, ref_logprobs=ref_logprobs,
                                       non_score_reward=non_score_reward, train_stats=train_stats,
                                       kl_coef=kl_coef)
        stats = stats_to_np(stats)
        timing['time/ppo/calc_stats'] = time.time() - t

        self.kl_ctl.update(stats['objective/kl'], self.ppo_params['batch_size'])

        timing['time/ppo/total'] = time.time() - t0
        stats.update(timing)
        return stats

    def batched_forward_pass(self, input_ids, attention_mask, encoder_hidden_states):
        """Calculate model outputs in multiple batches."""
        fbs = self.ppo_params['forward_batch_size']
        bs = input_ids.shape[0]

        logprobs = []
        ref_logprobs = []
        values = []
        for i in range(int(bs / min(fbs, bs))):
            m_input = input_ids[i * fbs:(i + 1) * fbs]
            m_mask = attention_mask[i * fbs:(i + 1) * fbs]
            encoder = encoder_hidden_states[i * fbs:(i + 1) * fbs]
            outputs = self.model(m_input,
                                 attention_mask=m_mask,
                                 encoder_outputs=encoder,
                                 output_hidden_states=True)
            last_hidden_state = outputs["hidden_states"][-1]
            logits = outputs["logits"]
            v = self.v_head(last_hidden_state).squeeze(-1)

            ref_logits = self.ref_model(m_input,
                                        attention_mask=m_mask,
                                        encoder_outputs=encoder)["logits"]

            values.append(v[:, :-1].detach())
            logprobs.append(logprobs_from_logits(logits[:, :-1, :], m_input[:, 1:]).detach())
            ref_logprobs.append(logprobs_from_logits(ref_logits[:, :-1, :], m_input[:, 1:]).detach())

        return torch.cat(logprobs), torch.cat(ref_logprobs), torch.cat(values)

    def train_minibatch(self, seq_len, logprobs, values, rewards, input_ids, attention_mask, images):
        """Train one PPO minibatch"""
        loss_p, loss_v, train_stats = self.loss(seq_len, logprobs, values, rewards, input_ids, attention_mask,
                                                images)
        loss = loss_p + loss_v
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        return train_stats

    def loss(self, seq_len, old_logprobs, values, rewards, input_ids, attention_mask, images):
        """Calculate policy and value losses."""
        lastgaelam = 0
        advantages_reversed = []
        gen_len = seq_len - 1  # we shifted inputs / logits
        for t in reversed(range(gen_len)):
            nextvalues = values[:, t + 1] if t < gen_len - 1 else 0.0
            delta = rewards[:, t] + self.ppo_params['gamma'] * nextvalues - values[:, t]
            lastgaelam = delta + self.ppo_params['gamma'] * self.ppo_params['lam'] * lastgaelam
            advantages_reversed.append(lastgaelam)
        advantages = torch.stack(advantages_reversed[::-1]).transpose(0, 1)

        returns = advantages + values
        advantages = whiten(advantages)
        advantages = advantages.detach()
        outputs = self.model(input_ids,
                             attention_mask=attention_mask,
                             encoder_outputs=self.encoder(images),
                             output_hidden_states=True)
        last_hidden_state = outputs["hidden_states"][-1]
        logits = outputs["logits"]
        vpred = self.v_head(last_hidden_state).squeeze(-1)
        vpred = vpred[:, :-1]
        logprob = logprobs_from_logits(logits[:, :-1, :], input_ids[:, 1:])

        vpredclipped = clip_by_value(vpred,
                                     values - self.ppo_params["cliprange_value"],
                                     values + self.ppo_params["cliprange_value"])

        vf_losses1 = (vpred - returns) ** 2
        vf_losses2 = (vpredclipped - returns) ** 2
        vf_loss = .5 * torch.mean(torch.max(vf_losses1, vf_losses2))
        vf_clipfrac = torch.mean(torch.gt(vf_losses2, vf_losses1).double())

        ratio = torch.exp(logprob - old_logprobs)

        pg_losses = -advantages * ratio
        pg_losses2 = -advantages * torch.clamp(ratio,
                                               1.0 - self.ppo_params['cliprange'],
                                               1.0 + self.ppo_params['cliprange'])

        pg_loss = torch.mean(torch.max(pg_losses, pg_losses2))
        pg_clipfrac = torch.mean(torch.gt(pg_losses2, pg_losses).double())

        loss = pg_loss + self.ppo_params['vf_coef'] * vf_loss

        entropy = torch.mean(entropy_from_logits(logits))
        approxkl = .5 * torch.mean((logprob - old_logprobs) ** 2)
        policykl = torch.mean(logprob - old_logprobs)
        return_mean, return_var = torch.mean(returns), torch.var(returns)
        value_mean, value_var = torch.mean(values), torch.var(values)

        stats = dict(
            loss=dict(policy=pg_loss, value=vf_loss, total=loss),
            policy=dict(entropy=entropy, approxkl=approxkl, policykl=policykl, clipfrac=pg_clipfrac,
                        advantages=advantages, advantages_mean=torch.mean(advantages), ratio=ratio),
            returns=dict(mean=return_mean, var=return_var),
            val=dict(vpred=torch.mean(vpred), error=torch.mean((vpred - returns) ** 2),
                     clipfrac=vf_clipfrac, mean=value_mean, var=value_var),
        )
        return pg_loss, self.ppo_params['vf_coef'] * vf_loss, flatten_dict(stats)

    def compute_rewards(self, scores, logprobs, ref_logprobs):
        """Compute per token rewards from scores and KL-penalty."""

        kl = logprobs - ref_logprobs
        non_score_reward = -self.kl_ctl.value * kl
        rewards = non_score_reward.clone().detach()
        rewards[np.arange(rewards.shape[0]), (rewards != 0).cumsum(1).argmax(1)] += scores.cuda()
        return rewards, non_score_reward, self.kl_ctl.value

    def get_stats(self, kl_coef, **data):

        stats = self.record_step_stats(kl_coef=kl_coef, **data)
        stats = stats_to_np(stats)
        self.stats['objective/kl'].append(stats["objective/kl"])
        self.stats['objective/kl_coef'] = stats["objective/kl_coef"]
        self.stats['objective/score'].append(stats["objective/score"])
        self.stats['ppo/mean_non_score_reward'].append(stats["ppo/mean_non_score_reward"])

        return self.stats

    def record_step_stats(self, kl_coef, **data):
        """Record training step statistics."""
        kl = data['logprobs'] - data['ref_logprobs']
        mean_kl = torch.mean(torch.sum(kl, axis=-1))
        mean_entropy = torch.mean(torch.sum(-data['logprobs'], axis=1))
        mean_non_score_reward = torch.mean(torch.sum(data['non_score_reward'], axis=1))
        score = torch.mean(data['scores'])
        stats = {
            'objective/kl': mean_kl,
            'objective/kl_dist': kl,
            'objective/score': score,
            'objective/logprobs': data['logprobs'],
            'objective/ref_logprobs': data['ref_logprobs'],
            'objective/kl_coef': kl_coef,
            'objective/entropy': mean_entropy,
            'ppo/mean_non_score_reward': mean_non_score_reward,
        }

        for k, v in data['train_stats'].items():
            stats[f'ppo/{k}'] = torch.mean(v, axis=0)
        stats['ppo/val/var_explained'] = 1 - stats['ppo/val/error'] / stats['ppo/returns/var']
        return stats



from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts


class DecreasingCosineAnnealingWarmRestarts(CosineAnnealingWarmRestarts):
    def __init__(self, factor, epochs, min_lr=0, eps=1e-8, **kwargs):
        super(DecreasingCosineAnnealingWarmRestarts, self).__init__(**kwargs)
        self.factor = factor
        self.epochs = epochs
        self.eps = eps
        self.min_lrs = [min_lr] * len(self.optimizer.param_groups)
        self.current_epoch = 0

    def step(self, epoch=None):
        super().step(epoch=epoch)
        if self.T_cur == 0:
            self.current_epoch += 1
        if self.current_epoch in self.epochs:
            self._reduce_lr()

    def _reduce_lr(self):
        for i, param_group in enumerate(self.optimizer.param_groups):
            old_lr = float(param_group['lr'])
            new_lr = max(old_lr * self.factor, self.min_lrs[i])
            if old_lr - new_lr > self.eps:
                param_group['lr'] = new_lr



import torch
import copy
import torch.nn as nn
from vilmedic.models.utils import get_n_params
from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.decoder.decoder_model import DecoderModel
from vilmedic.blocks.huggingface.encoder.encoder_model import EncoderModel
from vilmedic.blocks.huggingface.decoder.evaluation import evaluation
from vilmedic.models.rrg.RRG import RRG
from einops import rearrange


class RRG_MULTI(nn.Module):

    def __init__(self, encoder, decoder, cnn, dl, **kwargs):
        super().__init__()

        # Classical RRG
        decoder.vocab_size = dl.dataset.seq2seq.tgt.tokenizer.vocab_size
        self.mono_rrg = RRG(decoder, cnn, dl=None, **kwargs)

        # Linguistic encoder
        encoder.vocab_size = dl.dataset.seq2seq.src.tokenizer.vocab_size
        self.ling_enc = EncoderModel(encoder)

        # Evaluation
        self.eval_func = evaluation
        self.dec = self.mono_rrg.dec

    def forward(self,
                input_ids, attention_mask,
                decoder_input_ids, decoder_attention_mask,
                images, images_mask=None,
                encoder_outputs=None, encoder_attention_mask=None,
                epoch=None, iteration=None, **kwargs):

        if torch.cuda.is_available():
            input_ids = input_ids.cuda()
            attention_mask = attention_mask.cuda()
            decoder_input_ids = decoder_input_ids.cuda()
            decoder_attention_mask = decoder_attention_mask.cuda()

        if encoder_outputs is None:
            encoder_outputs, encoder_attention_mask = self.encode(input_ids,
                                                                  attention_mask,
                                                                  images,
                                                                  images_mask,
                                                                  **kwargs)

        # print(decoder_input_ids.shape)
        # print(decoder_attention_mask.shape)
        # print(encoder_outputs.shape)
        # print(encoder_attention_mask.shape)
        # troll
        out = self.mono_rrg.dec(input_ids=decoder_input_ids,
                                attention_mask=decoder_attention_mask,
                                encoder_outputs=encoder_outputs,
                                encoder_attention_mask=encoder_attention_mask,
                                **kwargs)

        return out

    # Necessary for generation
    def encode(self, input_ids, attention_mask, images, images_mask=None, **kwargs):
        # Visual features
        visual_features, visual_mask = self.mono_rrg.encode(images, images_mask=images_mask)

        # Linguistic features
        encoder_outputs = self.ling_enc(input_ids, attention_mask, return_dict=True)

        # Concatenation of features and masks
        encoder_hidden_states = torch.cat((encoder_outputs.last_hidden_state, visual_features), dim=1)
        encoder_attention_mask = torch.cat((attention_mask, visual_mask), dim=-1)

        return encoder_hidden_states, encoder_attention_mask

    def __repr__(self):
        s = "model: RRG\n"
        s += "(cnn):" + str(self.mono_rrg.enc) + '\n'
        s += "(enc):" + str(self.ling_enc) + '\n'
        s += "(dec):" + str(self.mono_rrg.dec) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.vision import CNN
from vilmedic.blocks.losses import VICREGLoss

from vilmedic.blocks.huggingface.encoder.encoder_model import EncoderModel

from tqdm import tqdm
import numpy as np


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]

    losses = []
    linguistics = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        batch = {k: v.cuda() if isinstance(v, torch.Tensor) else v for k, v in batch.items()}
        out = model(**batch)
        losses.append(out['loss'].mean().cpu().data.numpy())

        if not from_training:
            linguistics.append(out['linguistic'].cpu())
            visuals.append(out['visual'].cpu())

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'linguistic': torch.cat(linguistics),
            'visual': torch.cat(visuals)
            }


class VICREG(nn.Module):

    def __init__(self, encoder, cnn, projection, loss, **kwargs):
        super().__init__()

        # Linguistic encoder
        self.linguistic = EncoderModel(encoder)

        # Visual Encoder
        self.visual = eval(cnn.pop('proto'))(**cnn)

        # Projection
        self.vis_proj = nn.Sequential(
            nn.Linear(projection.visual_embedding_dim, projection.proj_hidden_dim),
            nn.BatchNorm1d(projection.proj_hidden_dim),
            nn.ReLU(),
            nn.Linear(projection.proj_hidden_dim, projection.proj_hidden_dim),
            nn.BatchNorm1d(projection.proj_hidden_dim),
            nn.ReLU(),
            nn.Linear(projection.proj_hidden_dim, projection.proj_output_dim),
        )

        self.lin_proj = nn.Sequential(
            nn.Linear(projection.textual_embedding_dim, projection.proj_hidden_dim),
            nn.BatchNorm1d(projection.proj_hidden_dim),
            nn.ReLU(),
            nn.Linear(projection.proj_hidden_dim, projection.proj_hidden_dim),
            nn.BatchNorm1d(projection.proj_hidden_dim),
            nn.ReLU(),
            nn.Linear(projection.proj_hidden_dim, projection.proj_output_dim),
        )

        self.loss_fn = VICREGLoss(**loss)

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, **kwargs):
        images = images.cuda()
        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()

        linguistic = self.linguistic(input_ids=input_ids,
                                     attention_mask=attention_mask,
                                     **kwargs)
        linguistic = self.lin_proj(linguistic['pooler_output'])
        visual = self.vis_proj(self.visual(images))

        vicreg_loss = self.loss_fn(linguistic, visual)

        return {"loss": vicreg_loss, "linguistic": linguistic, "visual": visual}

    def __repr__(self):
        s = "VICREG\n"
        s += str(self.visual) + '\n'
        s += str(self.linguistic) + '\n'
        s += str(self.lin_proj) + '\n'
        s += str(self.loss_fn) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



import logging
import functools
import torch.nn.functional as F
import inspect
import random
import torch.nn as nn
import torch
from collections import OrderedDict

from vilmedic.blocks.huggingface.encoder_decoder.evaluation import evaluation
from vilmedic.blocks.huggingface.encoder_decoder.encoder_decoder_model import EncoderDecoderModel
from vilmedic.blocks.scorers import Rouge
from vilmedic.blocks.scorers import CheXbert
from transformers.models.roberta.modeling_roberta import RobertaForCausalLM
import numpy as np


def create_state_dict(ckpt, diff, replace):
    weights = torch.load(ckpt)["model"]
    new_dict = set()
    for k, v in weights.items():
        if diff in k:  # avoid cnn
            new_dict.add((k.replace(diff, replace), v))
    new_dict = OrderedDict(new_dict)
    return new_dict


def to_contiguous(tensor):
    if tensor.is_contiguous():
        return tensor
    else:
        return tensor.contiguous()


def prepare_inputs_for_generation(self, input_ids, past=None, attention_mask=None, **model_kwargs):
    input_shape = input_ids.shape
    # if model is used as a decoder in encoder-decoder model, the decoder attention mask is created on the fly
    if attention_mask is None:
        attention_mask = input_ids.new_ones(input_shape)

    # cut decoder_input_ids if past is used
    if past is not None:
        input_ids = input_ids[:, -1:]

    return {"input_ids": input_ids, "attention_mask": attention_mask, "past_key_values": past, **model_kwargs}


def scst_loss(input, seq, reward, pad_token_id):
    # HuggingFace TopKLogitsWarper (if top_k > 0) puts -float('inf') for non top_k (or if we use bad_words_ids)
    # Padding can then have logits -float('inf') though we have to pad because hyp generation is finished
    # -float('inf') * 0 masking results in NaN so that doesnt mitigate the issue. need to do:
    input[input == -float("Inf")] = 0.
    #####

    # reward
    N, L = input.shape[:2]
    reward = reward.view(N, 1, 1).expand_as(input)
    input = -input * reward

    # masking
    mask = (seq > pad_token_id).float()
    output = input.view(-1) * mask.view(-1)

    # mean
    output = torch.sum(output) / torch.sum(mask)
    return output


class SumHugMono_SCST(nn.Module):

    def __init__(self, encoder, decoder, dl, logger=None, score="chexbert", top_k=None, **kwargs):
        super().__init__()
        self.enc_dec = EncoderDecoderModel(encoder, decoder)

        # Do we have pretrained ?
        if 'ckpt' in kwargs and 'enc' in kwargs['ckpt']:
            st = create_state_dict(kwargs['ckpt']['enc'], diff='linguistic.encoder.', replace='')
            self.enc_dec.enc_dec.encoder.load_state_dict(st, strict=True)
            print("loaded")
        elif 'ckpt' in kwargs and 'dec' in kwargs['ckpt']:
            st = create_state_dict(kwargs['ckpt']['dec'], diff='linguistic.encoder.', replace='roberta.')
            self.enc_dec.enc_dec.decoder.load_state_dict(st, strict=False)
        elif 'ckpt' in kwargs:
            self.load_state_dict(torch.load(kwargs["ckpt"])["model"], strict=True)

        self.score = score
        self.top_k = top_k
        # Scoring
        if score == "chexbert":
            self.chexbert = CheXbert()

        # Tokens
        self.bos_token_id = self.enc_dec.enc_dec.decoder.config.bos_token_id
        self.eos_token_id = self.enc_dec.enc_dec.decoder.config.eos_token_id
        self.pad_token_id = self.enc_dec.enc_dec.decoder.config.pad_token_id

        self.enc_dec.enc_dec.decoder.prepare_inputs_for_generation = functools.partial(prepare_inputs_for_generation,
                                                                                       self.enc_dec.enc_dec.decoder)

        # Tokenizer
        self.dl = dl
        self.tokenizer = dl.dataset.tgt_tokenizer
        self.logger = logger or logging.getLogger(__name__)

        # Evaluation
        self.eval_func = evaluation

        self.train_rewards = []

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask, **kwargs):
        batch_size = input_ids.shape[0]
        seq_len = input_ids.shape[1]

        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()
        decoder_input_ids = decoder_input_ids.cuda()

        # 1 Greedy
        with torch.no_grad():
            self.enc_dec.eval()
            encoder_outputs = self.enc_dec.enc_dec.encoder(input_ids, attention_mask)
            hidden_states = encoder_outputs[0]
            out = self.enc_dec.enc_dec.decoder.generate(
                input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
                max_length=seq_len,
                num_beams=1,
                num_return_sequences=1,
                return_dict_in_generate=True,
                output_scores=True,
                encoder_hidden_states=hidden_states,
                encoder_attention_mask=attention_mask,
                forced_eos_token_id=True,
            )
            greedy_input_ids = out.sequences
            score_greedy = self.get_reward(greedy_input_ids.data, decoder_input_ids)

        # 2. Sampling
        # self.enc_dec.train()
        self.enc_dec.eval()
        encoder_outputs = self.enc_dec.enc_dec.encoder(input_ids, attention_mask)
        hidden_states = encoder_outputs[0]
        out = inspect.unwrap(self.enc_dec.enc_dec.decoder.generate)(
            # inspect.unwrap removes the torch.no_grad() decorator
            self=self.enc_dec.enc_dec.decoder,
            do_sample=True,
            top_k=self.top_k,
            input_ids=torch.ones((batch_size, 1), dtype=torch.long).cuda() * self.bos_token_id,
            max_length=seq_len,
            num_beams=1,
            num_return_sequences=1,
            output_scores=True,
            return_dict_in_generate=True,
            encoder_hidden_states=hidden_states,
            encoder_attention_mask=attention_mask,
            bad_words_ids=[[self.pad_token_id], [self.bos_token_id]],
            forced_eos_token_id=True,
        )
        samples_ids = out.sequences[:, 1:]
        logits = torch.stack(out.scores, dim=1)
        logits = F.log_softmax(logits, dim=-1)
        sampled_logits = logits.gather(2, samples_ids.unsqueeze(-1))

        # 3. Reward and loss
        score_sampling = self.get_reward(samples_ids.data, decoder_input_ids)
        reward = torch.tensor(score_sampling).cuda() - torch.tensor(score_greedy).cuda()
        loss = scst_loss(sampled_logits, samples_ids.data, reward, self.pad_token_id)

        self.train_rewards.append(score_greedy)
        self.logger.info(np.mean(np.array(self.train_rewards)[-500:]))
        if random.randint(0, 500) == 0:
            self.logger.info("#" * 10)
            self.logger.info(
                self.tokenizer.decode(greedy_input_ids.data[0], skip_special_tokens=True,
                                      clean_up_tokenization_spaces=False))
            self.logger.info(
                self.tokenizer.decode(samples_ids.data[0], skip_special_tokens=True,
                                      clean_up_tokenization_spaces=False))
            self.logger.info("#" * 10)
        return {"loss": loss}

    def get_reward(self, rollout_input_ids, input_ids):

        hyp_list = []
        ref_list = []
        for h, r in zip(rollout_input_ids, input_ids):
            hyp_list.append(self.tokenizer.decode(h, skip_special_tokens=True, clean_up_tokenization_spaces=False))
            ref_list.append(self.tokenizer.decode(r, skip_special_tokens=True, clean_up_tokenization_spaces=False))
        if self.score == "chexbert":
            assert not self.chexbert.training
            chexbert, _ = self.chexbert(hyp_list, ref_list)
            reward = torch.tensor(chexbert["micro avg"]["f1-score"]).cuda()
        else:
            reward = Rouge(rouges=[self.score]).compute(ref_list, hyp_list)[1]
        return reward

    def __repr__(self):
        return "SumHugMono\n" + str(self.enc_dec)



import torch
import torch.nn as nn
from random import sample
import torchvision.models as models


class MoCo(nn.Module):
    """
    Build a MoCo model with: a query encoder, a key encoder, and a queue
    https://arxiv.org/abs/1911.05722
    """

    def __init__(self, base_encoder, dim=128, r=16384, m=0.999, T=0.1, mlp=False):
        """
        dim: feature dimension (default: 128)
        r: queue size; number of negative samples/prototypes (default: 16384)
        m: momentum for updating key encoder (default: 0.999)
        T: softmax temperature
        mlp: whether to use mlp projection
        """
        super(MoCo, self).__init__()

        base_encoder = models.__dict__[base_encoder]
        self.r = r
        self.m = m
        self.T = T

        # create the encoders
        # num_classes is the output fc dimension
        self.encoder_q = base_encoder(num_classes=dim)
        self.encoder_k = base_encoder(num_classes=dim)

        if mlp:  # hack: brute-force replacement
            dim_mlp = self.encoder_q.fc.weight.shape[1]
            self.encoder_q.fc = nn.Sequential(nn.Linear(dim_mlp, dim_mlp), nn.ReLU(), self.encoder_q.fc)
            self.encoder_k.fc = nn.Sequential(nn.Linear(dim_mlp, dim_mlp), nn.ReLU(), self.encoder_k.fc)

        for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
            param_k.data.copy_(param_q.data)  # initialize
            param_k.requires_grad = False  # not update by gradient

        # create the queue
        self.register_buffer("queue", torch.randn(r, dim))
        self.queue = nn.functional.normalize(self.queue, dim=0)

        self.register_buffer("queue_ptr", torch.zeros(1, dtype=torch.long))

    @torch.no_grad()
    def _momentum_update_key_encoder(self):
        """
        Momentum update of the key encoder
        """
        for param_q, param_k in zip(self.encoder_q.parameters(), self.encoder_k.parameters()):
            param_k.data = param_k.data * self.m + param_q.data * (1. - self.m)

    @torch.no_grad()
    def _dequeue_and_enqueue(self, keys):
        # gather keys before updating queue

        batch_size = keys.shape[0]

        ptr = int(self.queue_ptr)
        assert self.r % batch_size == 0  # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.queue[ptr:ptr + batch_size, :] = keys
        ptr = (ptr + batch_size) % self.r  # move pointer

        self.queue_ptr[0] = ptr

    def forward(self, im_q, im_k=None, is_eval=False, cluster_result=None, index=None):
        """
        Input:
            im_q: a batch of query images
            im_k: a batch of key images
            is_eval: return momentum embeddings (used for clustering)
            cluster_result: cluster assignments, centroids, and density
            index: indices for training samples
        Output:
            logits, targets, proto_logits, proto_targets
        """

        if is_eval:
            k = self.encoder_k(im_q)
            k = nn.functional.normalize(k, dim=1)
            return k

        # compute key features
        with torch.no_grad():  # no gradient to keys
            self._momentum_update_key_encoder()  # update the key encoder
            k = self.encoder_k(im_k)  # keys: NxC
            k = nn.functional.normalize(k, dim=1)

        # compute query features
        q = self.encoder_q(im_q)  # queries: NxC
        q = nn.functional.normalize(q, dim=1)

        logits = q @ torch.cat([k, self.queue]).T / self.T  # Nx(N+K) instead of original Nx(1+K)
        labels = torch.arange(len(q)).cuda()

        # dequeue and enqueue
        self._dequeue_and_enqueue(k)
        # prototypical contrast
        if cluster_result is not None:
            proto_labels = []
            proto_logits = []
            for n, (im2cluster, prototypes, density) in enumerate(
                    zip(cluster_result['im2cluster'], cluster_result['centroids'], cluster_result['density'])):
                # get positive prototypes
                pos_proto_id = im2cluster[index]
                pos_prototypes = prototypes[pos_proto_id]

                # sample negative prototypes
                all_proto_id = [i for i in range(im2cluster.max())]
                neg_proto_id = set(all_proto_id) - set(pos_proto_id.tolist())
                neg_proto_id = sample(neg_proto_id, self.r)  # sample r negative prototypes
                neg_prototypes = prototypes[neg_proto_id]

                proto_selected = torch.cat([pos_prototypes, neg_prototypes], dim=0)

                # compute prototypical logits
                logits_proto = torch.mm(q, proto_selected.t())

                # targets for prototype assignment
                labels_proto = torch.linspace(0, q.size(0) - 1, steps=q.size(0)).long().cuda()

                # scaling temperatures for the selected prototypes
                temp_proto = density[torch.cat([pos_proto_id, torch.LongTensor(neg_proto_id).cuda()], dim=0)]
                logits_proto /= temp_proto

                proto_labels.append(labels_proto)
                proto_logits.append(logits_proto)
            return logits, labels, proto_logits, proto_labels
        else:
            return logits, labels, None, None



import math
import warnings
from typing import List

from torch import nn
from torch.optim import Adam, Optimizer
from torch.optim.lr_scheduler import _LRScheduler


# https://raw.githubusercontent.com/PyTorchLightning/lightning-bolts/0.5.0/pl_bolts/optimizers/lr_scheduler.py
class LinearWarmupCosineAnnealingLR(_LRScheduler):
    """Sets the learning rate of each parameter group to follow a linear warmup schedule between warmup_start_lr
    and base_lr followed by a cosine annealing schedule between base_lr and eta_min.

    .. warning::
        It is recommended to call :func:`.step()` for :class:`LinearWarmupCosineAnnealingLR`
        after each iteration as calling it after each epoch will keep the starting lr at
        warmup_start_lr for the first epoch which is 0 in most cases.

    .. warning::
        passing epoch to :func:`.step()` is being deprecated and comes with an EPOCH_DEPRECATION_WARNING.
        It calls the :func:`_get_closed_form_lr()` method for this scheduler instead of
        :func:`get_lr()`. Though this does not change the behavior of the scheduler, when passing
        epoch param to :func:`.step()`, the user should call the :func:`.step()` function before calling
        train and validation methods.

    Example:
        >>> layer = nn.Linear(10, 1)
        >>> optimizer = Adam(layer.parameters(), lr=0.02)
        >>> scheduler = LinearWarmupCosineAnnealingLR(optimizer, warmup_epochs=10, max_epochs=40)
        >>> #
        >>> # the default case
        >>> for epoch in range(40):
        ...     # train(...)
        ...     # validate(...)
        ...     scheduler.step()
        >>> #
        >>> # passing epoch param case
        >>> for epoch in range(40):
        ...     scheduler.step(epoch)
        ...     # train(...)
        ...     # validate(...)
    """

    def __init__(
            self,
            optimizer: Optimizer,
            warmup_epochs: int,
            max_epochs: int,
            warmup_start_lr: float = 0.0,
            eta_min: float = 0.0,
            last_epoch: int = -1,
    ) -> None:
        """
        Args:
            optimizer (Optimizer): Wrapped optimizer.
            warmup_epochs (int): Maximum number of iterations for linear warmup
            max_epochs (int): Maximum number of iterations
            warmup_start_lr (float): Learning rate to start the linear warmup. Default: 0.
            eta_min (float): Minimum learning rate. Default: 0.
            last_epoch (int): The index of last epoch. Default: -1.
        """
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs
        self.warmup_start_lr = warmup_start_lr
        self.eta_min = eta_min

        super().__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:
        """Compute learning rate using chainable form of the scheduler."""
        if not self._get_lr_called_within_step:
            warnings.warn(
                "To get the last learning rate computed by the scheduler, " "please use `get_last_lr()`.",
                UserWarning,
            )

        if self.last_epoch == 0:
            return [self.warmup_start_lr] * len(self.base_lrs)
        if self.last_epoch < self.warmup_epochs:
            return [
                group["lr"] + (base_lr - self.warmup_start_lr) / (self.warmup_epochs - 1)
                for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)
            ]
        if self.last_epoch == self.warmup_epochs:
            return self.base_lrs
        if (self.last_epoch - 1 - self.max_epochs) % (2 * (self.max_epochs - self.warmup_epochs)) == 0:
            return [
                group["lr"]
                + (base_lr - self.eta_min) * (1 - math.cos(math.pi / (self.max_epochs - self.warmup_epochs))) / 2
                for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)
            ]

        return [
            (1 + math.cos(math.pi * (self.last_epoch - self.warmup_epochs) / (self.max_epochs - self.warmup_epochs)))
            / (
                    1
                    + math.cos(
                math.pi * (self.last_epoch - self.warmup_epochs - 1) / (self.max_epochs - self.warmup_epochs)
            )
            )
            * (group["lr"] - self.eta_min)
            + self.eta_min
            for group in self.optimizer.param_groups
        ]

    def _get_closed_form_lr(self) -> List[float]:
        """Called when epoch is passed as a param to the `step` function of the scheduler."""
        if self.last_epoch < self.warmup_epochs:
            return [
                self.warmup_start_lr + self.last_epoch * (base_lr - self.warmup_start_lr) / (self.warmup_epochs - 1)
                for base_lr in self.base_lrs
            ]

        return [
            self.eta_min
            + 0.5
            * (base_lr - self.eta_min)
            * (1 + math.cos(math.pi * (self.last_epoch - self.warmup_epochs) / (self.max_epochs - self.warmup_epochs)))
            for base_lr in self.base_lrs
        ]


# warmup + decay as a function
def linear_warmup_decay(warmup_steps, total_steps, cosine=True, linear=False):
    """Linear warmup for warmup_steps, optionally with cosine annealing or linear decay to 0 at total_steps."""
    assert not (linear and cosine)

    def fn(step):
        if step < warmup_steps:
            return float(step) / float(max(1, warmup_steps))

        if not (cosine or linear):
            # no decay
            return 1.0

        progress = float(step - warmup_steps) / float(max(1, total_steps - warmup_steps))
        if cosine:
            # cosine decay
            return 0.5 * (1.0 + math.cos(math.pi * progress))

        # linear decay
        return 1.0 - progress

    return fn



#!/usr/bin/env python
# 
# File Name : bleu.py
#
# Description : Wrapper for BLEU scorer.
#
# Creation Date : 06-01-2015
# Last Modified : Thu 19 Mar 2015 09:13:28 PM PDT
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>

import torch.nn as nn
from .bleu_scorer import BleuScorer


class Bleu(nn.Module):
    def __init__(self, n=4, **kwargs):
        # default compute Blue score up to 4
        super().__init__()
        self._n = n

    def forward(self, gts, res):
        return self.compute_score(gts, res)

    def compute_score(self, gts, res):
        res = {i: [v] for i, v in enumerate(res)}
        gts = {i: [v] for i, v in enumerate(gts)}
        bleu_scorer = BleuScorer(n=self._n)

        for id in sorted(gts.keys()):
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert (type(hypo) is list)
            assert (len(hypo) == 1)
            assert (type(ref) is list)
            assert (len(ref) >= 1)

            bleu_scorer += (hypo[0], ref)

        # score, scores = bleu_scorer.compute_score(option='shortest')
        score, scores = bleu_scorer.compute_score(option='closest', verbose=0)
        # score, scores = bleu_scorer.compute_score(option='average', verbose=1)

        # return (bleu, bleu_info)
        return score[self._n-1], scores[self._n-1]

    def method(self):
        return "Bleu"



from vilmedic.blocks.others.mcan import MCA_ED
from vilmedic.blocks.others.mcan import LayerNorm
from vilmedic.blocks.others.mcan.make_mask import make_mask
from vilmedic.blocks.others.mcan import AttFlat

from vilmedic.blocks.classifier.evaluation import evaluation
from vilmedic.blocks.classifier.losses import get_loss
import torch.nn as nn


class SIM_MCAN(nn.Module):
    def __init__(self, visual, adapter, linguistic, answer_size, loss, **kwargs):
        super(SIM_MCAN, self).__init__()

        self.embedding = nn.Embedding(
            num_embeddings=linguistic.TOKEN_SIZE,
            embedding_dim=linguistic.WORD_EMBED_SIZE
        )

        visual_func = visual.pop('proto')
        self.cnn = eval(visual_func)(**visual)

        self.adapter = nn.Sequential(
            nn.Linear(adapter.pop('input_size'), adapter.pop('output_size')),
            # torch.nn.LayerNorm(linguistic.hidden_size, eps=linguistic.layer_norm_eps)
        )

        # # Loading the GloVe embedding weights
        # if __C.USE_GLOVE:
        #     self.embedding.weight.data.copy_(torch.from_numpy(pretrained_emb))

        self.lstm = nn.LSTM(
            input_size=linguistic.WORD_EMBED_SIZE,
            hidden_size=linguistic.HIDDEN_SIZE,
            num_layers=1,
            batch_first=True
        )

        self.backbone = MCA_ED(linguistic)

        # Flatten to vector
        self.attflat_img = AttFlat(linguistic)
        self.attflat_lang = AttFlat(linguistic)

        # Classification layers
        self.proj_norm = LayerNorm(linguistic.FLAT_OUT_SIZE)
        self.proj = nn.Linear(linguistic.FLAT_OUT_SIZE, answer_size)

        self.loss_func = get_loss(loss.pop('proto'), **loss).cuda()
        # Evaluation
        self.eval_func = evaluation

    def forward(self, images, input_ids, attention_mask, labels, **kwargs):
        # Pre-process Language Feature
        input_ids = input_ids.cuda()
        images = images.cuda()

        img_feat = self.cnn(images)

        img_feat = self.adapter(img_feat)
        lang_feat_mask = make_mask(input_ids.unsqueeze(2))
        lang_feat = self.embedding(input_ids)
        lang_feat, _ = self.lstm(lang_feat)

        img_feat_mask = make_mask(img_feat)

        # Backbone Framework
        lang_feat, img_feat = self.backbone(
            lang_feat,
            img_feat,
            lang_feat_mask,
            img_feat_mask
        )

        # Flatten to vector
        lang_feat = self.attflat_lang(
            lang_feat,
            lang_feat_mask
        )

        img_feat = self.attflat_img(
            img_feat,
            img_feat_mask
        )

        # Classification layers
        proj_feat = lang_feat + img_feat
        proj_feat = self.proj_norm(proj_feat)
        output = self.proj(proj_feat)

        loss = self.loss_func(output, labels.cuda(), **kwargs)
        
        return {'loss': loss, 'output': output}



from functools import wraps
import torch
from torch import nn

from vit_pytorch.vit import Attention

def find_modules(nn_module, type):
    return [module for module in nn_module.modules() if isinstance(module, type)]

class Recorder(nn.Module):
    def __init__(self, vit, device = None):
        super().__init__()
        self.vit = vit

        self.data = None
        self.recordings = []
        self.hooks = []
        self.hook_registered = False
        self.ejected = False
        self.device = device

    def _hook(self, _, input, output):
        self.recordings.append(output.clone().detach())

    def _register_hook(self):
        modules = find_modules(self.vit.transformer, Attention)
        for module in modules:
            handle = module.attend.register_forward_hook(self._hook)
            self.hooks.append(handle)
        self.hook_registered = True

    def eject(self):
        self.ejected = True
        for hook in self.hooks:
            hook.remove()
        self.hooks.clear()
        return self.vit

    def clear(self):
        self.recordings.clear()

    def record(self, attn):
        recording = attn.clone().detach()
        self.recordings.append(recording)

    def forward(self, img):
        assert not self.ejected, 'recorder has been ejected, cannot be used anymore'
        self.clear()
        if not self.hook_registered:
            self._register_hook()

        pred = self.vit(img)

        # move all recordings to one device before stacking
        target_device = self.device if self.device is not None else img.device
        recordings = tuple(map(lambda t: t.to(target_device), self.recordings))

        attns = torch.stack(recordings, dim = 1)
        return pred, attns



import numpy as np
# 
# nli_model = SimpleNLI(model, batch=self.nli_batch, neutral_score=self.nli_neutral_score,
#                                       nthreads=self.nthreads, pin_memory=self.pin_memory,
#                                       bert_score=self.BERT_SCORE_DEFAULT, sentsplitter='linebreak',
#                                       cache=self.nli_cache, verbose=self.verbose)


def score(self, rids, hypos):
    # Named entity recognition
    hypo_sents = {}
    hypos_entities = {}
    texts, buf = [], []
    for hypo in hypos:
        buf.append(hypo)
        if len(buf) >= self.batch:
            text = '\n\n{0}\n\n'.format(self.DOC_SEPARATOR).join(buf)
            texts.append(text)
            buf = []
    if len(buf) > 0:
        text = '\n\n{0}\n\n'.format(self.DOC_SEPARATOR).join(buf)
        texts.append(text)
    i = 0
    for text in texts:
        doc = self.ner(text)
        j = 0
        for sentence in doc.sentences:
            if i not in hypos_entities:
                hypos_entities[i] = {}
            if i not in hypo_sents:
                hypo_sents[i] = ''
            if sentence.text == self.DOC_SEPARATOR:
                i += 1
                j = 0
            else:
                if len(hypo_sents[i]) > 0:
                    hypo_sents[i] += '\n'
                hypo_sents[i] += sentence.text
                for entity in sentence.ents:
                    if entity.type in self.target_types:
                        buf = []
                        for word in entity.words:
                            buf.append(word.text.lower())
                        s = ' '.join(buf)
                        if s not in hypos_entities:
                            hypos_entities[i][s] = [j]
                        else:
                            hypos_entities[i][s].append(j)
                j += 1
        i += 1
    hypo_nli, ref_nli = None, None
    if self.mode.startswith(self.MODE_NLI):
        hypo_nli, ref_nli = {}, {}
        texts1, texts2 = [], []
        for i, rid in enumerate(rids):
            buf = []
            rid = rid.split(self.ID_SEPARATOR)[0]
            for sid in sorted(self.sentences[rid].keys()):
                buf.append(self.sentences[rid][sid])
            texts1.append('\n'.join(buf))
            texts2.append(hypo_sents[i])
        _, _, _, stats = self.nli.sentence_scores_bert_score(texts1, texts2, label='all', prf=self.prf)
        for i in range(len(rids)):
            rid, rs = rids[i], stats[i]
            ref_nli[i] = {}
            for sid, tup in rs['scores'][0].items():
                pred, _ = self._nli_label(tup[0])
                ref_nli[i][sid] = pred
            hypo_nli[i] = {}
            for sid, tup in rs['scores'][1].items():
                pred, _ = self._nli_label(tup[0])
                hypo_nli[i][sid] = pred
    # Calculate scores
    scores_e, scores_n = [], []
    for i, rid in enumerate(rids):
        hypo_entities = hypos_entities[i]
        rid = rid.split(self.ID_SEPARATOR)[0]
        ref_entities = self.entities[rid]
        # precision
        match_e, match_n, total_pr = 0, 0, 0
        if self.prf != 'r':
            for s in hypo_entities.keys():
                for sid in hypo_entities[s]:
                    if s in ref_entities:
                        match_e += 1
                        if hypo_nli is None:
                            match_n += 1.0
                    if hypo_nli is not None:
                        if hypo_nli[i][sid] == 'neutral':
                            if s in ref_entities:
                                match_n += 1.0
                        elif hypo_nli[i][sid] == 'entailment':
                            if s in hypo_entities:
                                match_n += 1.0
                            else:
                                if self.mode == self.MODE_NLI or self.mode == self.MODE_NLI_ENTAILMENT:
                                    match_n += self.entail_score
                        elif hypo_nli[i][sid] == 'contradiction':
                            if self.mode == self.MODE_NLI_ENTAILMENT:
                                if s in hypo_entities:
                                    match_n += 1.0
                    total_pr += 1
        pr_e = match_e / total_pr if total_pr > 0 else 0.0
        pr_n = match_n / total_pr if total_pr > 0 else 0.0
        # recall
        match_e, match_n, total_rc = 0, 0, 0
        if self.prf != 'p':
            for s in ref_entities.keys():
                for sid in ref_entities[s]:
                    if s in hypo_entities:
                        match_e += 1
                        if ref_nli is None:
                            match_n += 1.0
                    if ref_nli is not None:
                        if ref_nli[i][sid] == 'neutral':
                            if s in hypo_entities:
                                match_n += 1.0
                        elif ref_nli[i][sid] == 'entailment':
                            if s in hypo_entities:
                                match_n += 1.0
                            else:
                                if self.mode == self.MODE_NLI or self.mode == self.MODE_NLI_ENTAILMENT:
                                    match_n += self.entail_score
                        elif ref_nli[i][sid] == 'contradiction':
                            if self.mode == self.MODE_NLI_ENTAILMENT:
                                if s in hypo_entities:
                                    match_n += 1.0
                    total_rc += 1
        rc_e = match_e / total_rc if total_rc > 0 else 0.0
        rc_n = match_n / total_rc if total_rc > 0 else 0.0
        # fb1
        if self.prf == 'p':
            score_e, score_n = pr_e, pr_n
        elif self.prf == 'r':
            score_e, score_n = rc_e, rc_n
        else:
            score_e = 2 * pr_e * rc_e / (pr_e + rc_e) if pr_e > 0.0 and rc_e > 0.0 else 0.0
            score_n = 2 * pr_n * rc_n / (pr_n + rc_n) if pr_n > 0.0 and rc_n > 0.0 else 0.0
        if self.penalty:
            penalty = np.e ** (-((total_pr - total_rc) ** 2) / (2 * self.PENALTY_SIGMA ** 2))
            score_e *= penalty
            score_n *= penalty
        scores_e.append(score_e)
        scores_n.append(score_n)
    mean_exact_e = np.mean(scores_e)
    mean_exact_n = np.mean(scores_n)
    return mean_exact_e, scores_e, mean_exact_n, scores_n



import torch.nn as nn
import torch
import math
import torch.nn.functional as F

def get_activation_fn(name):
    """Returns a callable activation function from torch."""
    if name in (None, 'linear'):
        return lambda x: x
    elif name in ('sigmoid', 'tanh'):
        return getattr(torch, name)
    else:
        return getattr(F, name)


class FF(nn.Module):
    """A smart feedforward layer with activation support.
    Arguments:
        in_features(int): Input dimensionality.
        out_features(int): Output dimensionality.
        bias(bool, optional): Enable/disable bias for the layer. (Default: True)
        bias_zero(bool, optional): Start with a 0-vector bias. (Default: True)
        activ(str, optional): A string like 'tanh' or 'relu' to define the
            non-linearity type. `None` or `'linear'` is a linear layer (default).
    """

    def __init__(self, in_features, out_features, bias=True,
                 bias_zero=True, activ=None):
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.use_bias = bias
        self.bias_zero = bias_zero
        self.activ_type = activ
        if self.activ_type in (None, 'linear'):
            self.activ_type = 'linear'
        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        self.activ = get_activation_fn(activ)

        if self.use_bias:
            self.bias = nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.use_bias:
            if self.bias_zero:
                self.bias.data.zero_()
            else:
                self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input):
        return self.activ(F.linear(input, self.weight, self.bias))

    def __repr__(self):
        repr_ = self.__class__.__name__ + '(' \
            + 'in_features=' + str(self.in_features) \
            + ', out_features=' + str(self.out_features) \
            + ', activ=' + str(self.activ_type) \
            + ', bias=' + str(self.use_bias)
        if self.use_bias:
            repr_ += ', bias_zero=' + str(self.bias_zero)
        return repr_ + ')'


import gzip
import os
import site
import sys
import time
import numpy as np
import pandas as pd
import torch
from collections import defaultdict, OrderedDict
from bert_score.utils import bert_cos_score_idf, cache_scibert, get_idf_dict, get_model, lang2model, model2layers
from cachetools import LRUCache
from torch.nn import DataParallel
from torch.nn.functional import softmax
from transformers import AutoTokenizer
from nltk.tokenize import wordpunct_tokenize
from .BERTNLI import BERTNLI


class NLTKTokenizer:
    @staticmethod
    def tokenize(text):
        return wordpunct_tokenize(text)


class _NLIScorer:
    LABEL_ALL = 'all'
    LABEL_CONTRADICT = 'contradiction'
    LABEL_ENTAIL = 'entailment'
    LABEL_NEUTRAL = 'neutral'
    METRIC_FB1 = 'f'
    PENALTY_SIGMA = 6.0
    THRESH_FIX = 'fix'
    THRESH_NONE = 'none'

    def __init__(self, neutral_score=(1.0 / 3), batch=16, nthreads=2, pin_memory=False, bert_score=None,
                 cache=None, verbose=False):
        self.model = None
        self.neutral_score = neutral_score
        self.batch = batch
        self.nthreads = nthreads
        self.pin_memory = pin_memory

        self.tokenizer = NLTKTokenizer()
        if bert_score is not None:
            self.bert_score_model = BERTScorer(model_type=bert_score, batch_size=batch, nthreads=nthreads,
                                               lang='en', rescale_with_baseline=True)
        else:
            self.bert_score_model = None
        self.cache = LRUCache(cache) if cache is not None else None
        self.verbose = verbose
        self.gpu = False

    def cuda(self):
        self.model = self.model.cuda()
        self.gpu = True
        if self.bert_score_model is not None:
            self.bert_score_model = self.bert_score_model.cuda()
        return self

    def predict(self, premises, hypotheses):
        raise NotImplementedError

    def sentence_scores_bert_score(self, texts1, texts2, label='entailment', thresh='none', prf='f'):
        # Calculate BertScores
        tids, tsents1, tsents2, bsents1, bsents2 = [], {}, {}, [], []
        bsents1t, bsents2t, thresh_nums = [], [], [0, 0]
        for tid, (text1, text2) in enumerate(zip(texts1, texts2)):
            tids.append(tid)
            sents1 = text1
            tsents1[tid] = sents1
            sents2 = text2
            tsents2[tid] = sents2
            for i, sent1 in enumerate(sents1):
                for j, sent2 in enumerate(sents2):
                    bsents1.append(sent1)
                    bsents2.append(sent2)
        bsents1 += bsents1t
        bsents2 += bsents2t
        # Store BERTScores to dictionaries of (#text, #sentence)
        _, _, bf = self.bert_score_model.score(bsents1, bsents2)
        bf = bf.numpy()
        idx = 0
        scores1, scores2 = {}, {}
        for tid in tids:
            scores1[tid], scores2[tid] = {}, {}
            for i in range(len(tsents1[tid])):
                scores1[tid][i] = []
                for j in range(len(tsents2[tid])):
                    if j not in scores2[tid]:
                        scores2[tid][j] = []
                    score = bf[idx]
                    scores1[tid][i].append(score)
                    scores2[tid][j].append(score)
                    idx += 1
        thresh1, thresh2 = -100.0, -100.0
        # Obtain high BertScore premise-hypothesis pairs
        tids_all, prems_all, hypos_all = [], [], []
        prems, hypos, pbfs, fidxs = {}, {}, {}, {}
        valid_sents, cache_sents = {}, {}
        for k in [0, 1]:
            prems[k], hypos[k], pbfs[k], fidxs[k] = {}, {}, {}, {}
            valid_sents[k], cache_sents[k] = {}, {}
        for tid in tids:
            for k, quad in enumerate([(scores1, tsents1, tsents2, thresh1), (scores2, tsents2, tsents1, thresh2)]):
                # ref sent: prem=gen & hypo=ref, gen sent: prem=ref & hypo=gen
                scores, sents_hypos, sents_prems, thresh = quad
                prems[k][tid], hypos[k][tid], pbfs[k][tid], fidxs[k][tid] = {}, {}, {}, {}
                valid_sents[k][tid], cache_sents[k][tid] = {}, {}
                for i in range(len(scores[tid])):
                    if len(scores[tid][i]) > 0:
                        fidx = np.argmax(scores[tid][i])
                        fscore = scores[tid][i][fidx]
                        # Align with the best scoring sentence
                        prems[k][tid][i] = sents_prems[tid][fidx]
                        hypos[k][tid][i] = sents_hypos[tid][i]
                        pbfs[k][tid][i] = fscore
                        fidxs[k][tid][i] = fidx
                        # Set IDs, premises, and hypothesis for NLI
                        prem = sents_prems[tid][fidx]
                        hypo = sents_hypos[tid][i]
                        if self.cache is not None and (prem, hypo) in self.cache:
                            cache_sents[k][tid][i] = self.cache[(prem, hypo)]
                            valid_sents[k][tid][i] = None
                        else:
                            if scores[tid][i][fidx] >= thresh and ((k == 0 and prf != 'p') or (k == 1 and prf != 'r')):
                                tids_all.append(tid)
                                prems_all.append(prem)
                                hypos_all.append(hypo)
                                valid_sents[k][tid][i] = (prem, hypo)
                            else:
                                valid_sents[k][tid][i] = None

        # NLI probabilities
        probs_all, _ = self.predict(prems_all, hypos_all)
        probs_tids = OrderedDict()
        for i, tid in enumerate(tids_all):
            if tid not in probs_tids:
                probs_tids[tid] = []
            probs_tids[tid].append(probs_all[i])
        prs, rcs, fb1s, stats = [], [], [], []
        for tid in tids:
            rc_pr = {0: [], 1: []}
            sent_probs = [{}, {}]
            idx = 0

            for k, num_sent in zip([0, 1], (len(tsents1[tid]), len(tsents2[tid]))):
                for i in range(num_sent):
                    if i in cache_sents[k][tid]:
                        prob = cache_sents[k][tid][i]
                    elif i in valid_sents[k][tid] and valid_sents[k][tid][i] is not None:
                        prem_hypo = valid_sents[k][tid][i]
                        prob = probs_tids[tid][idx]
                        if self.cache is not None:
                            self.cache[prem_hypo] = prob
                        idx += 1
                    else:
                        prob = None

                    # Soft scoring
                    if prob is not None:
                        score = prob[self.LABEL_ENTAIL]
                        rc_pr[k].append(score)
                    else:
                        prob = {self.LABEL_ENTAIL: -1.0, self.LABEL_NEUTRAL: -1.0, self.LABEL_CONTRADICT: -1.0}
                    if i in prems[k][tid] and i in hypos[k][tid]:
                        sent_probs[k][i] = (prob, pbfs[k][tid][i], prems[k][tid][i], hypos[k][tid][i], fidxs[k][tid][i])
            mean_precision = np.mean(rc_pr[1]) if len(rc_pr[1]) > 0 else 0.0
            mean_recall = np.mean(rc_pr[0]) if len(rc_pr[0]) > 0 else 0.0
            if mean_precision + mean_recall > 0.0:
                fb1 = 2 * mean_precision * mean_recall / (mean_precision + mean_recall)
            else:
                fb1 = 0.0
            prs.append(mean_precision)
            rcs.append(mean_recall)
            fb1s.append(fb1)
            stats.append({'scores': sent_probs, 'threshes': (thresh1, thresh2)})
        return prs, rcs, fb1s, stats

    def stop(self):
        pass


class BERTScorer:
    PENALTY_SIGMA = 6.0

    def __init__(self, refs=None, model_type=None, num_layers=None, verbose=False, idf=False, batch_size=16, nthreads=2,
                 all_layers=False, lang=None, rescale_with_baseline=False, penalty=False):
        assert lang is not None or model_type is not None, 'Either lang or model_type should be specified'
        if rescale_with_baseline:
            assert lang is not None, 'Need to specify Language when rescaling with baseline'

        if model_type is None:
            lang = lang.lower()
            model_type = lang2model[lang]
        if num_layers is None:
            num_layers = model2layers[model_type]

        if model_type.startswith('scibert'):
            tokenizer = AutoTokenizer.from_pretrained(cache_scibert(model_type))
        else:
            tokenizer = AutoTokenizer.from_pretrained(model_type)
        model = get_model(model_type, num_layers, all_layers)

        if not idf:
            idf_dict = defaultdict(lambda: 1.)
            # set idf for [SEP] and [CLS] to 0
            idf_dict[tokenizer.sep_token_id] = 0
            idf_dict[tokenizer.cls_token_id] = 0
        elif isinstance(idf, dict):
            if verbose:
                print('using predefined IDF dict...')
            idf_dict = idf
        else:
            if verbose:
                print('preparing IDF dict...')
            start = time.perf_counter()
            idf_dict = get_idf_dict(refs, tokenizer, nthreads=nthreads)
            if verbose:
                print('done in {:.2f} seconds'.format(time.perf_counter() - start))

        self.batch_size = batch_size
        self.verbose = verbose
        self.all_layers = all_layers
        self.penalty = penalty

        self.tokenizer = tokenizer
        self.model = model
        self.idf_dict = idf_dict
        self.device = 'cpu'

        self.baselines = None
        if rescale_with_baseline:
            self.baselines = torch.tensor([0.6666, 0.6666, 0.6662])

    def cuda(self):
        self.device = 'cuda:0'
        self.model.cuda()
        return self

    def score(self, cands, refs):
        assert len(cands) == len(refs)

        if self.verbose:
            print('calculating scores...')
        start = time.perf_counter()
        all_preds = bert_cos_score_idf(self.model, refs, cands, self.tokenizer, self.idf_dict, verbose=self.verbose,
                                       device=self.device, batch_size=self.batch_size, all_layers=self.all_layers).cpu()
        if self.baselines is not None:
            all_preds = (all_preds - self.baselines) / (1 - self.baselines)

        out = all_preds[..., 0], all_preds[..., 1], all_preds[..., 2]  # P, R, F
        if self.penalty:
            for idx, (cand, ref) in enumerate(zip(cands, refs)):
                toks1 = self.tokenizer.tokenize(cand)
                toks2 = self.tokenizer.tokenize(ref)
                penalty = np.e ** (-((len(toks1) - len(toks2)) ** 2) / (2 * self.PENALTY_SIGMA ** 2))
                out[-1][idx] *= penalty

        if self.verbose:
            time_diff = time.perf_counter() - start
            print(f'done in {time_diff:.2f} seconds, {len(refs) / time_diff:.2f} sentences/sec')

        return out


class SimpleNLI(_NLIScorer):
    def __init__(self, model, neutral_score=(1.0 / 3), batch=16, nthreads=2, pin_memory=False, bert_score=None,
                 cache=None, verbose=False):
        super(SimpleNLI, self).__init__(neutral_score, batch, nthreads, pin_memory, bert_score, cache,
                                        verbose)

        self.model = DataParallel(model)
        self.model.cuda()
        self.model.eval()
        self.model.eval()

    @classmethod
    def load_model(cls, states=None):
        bert_type = 'bert'
        name = 'bert-base-uncased'
        bertnli = BERTNLI(name, bert_type=bert_type, length=384, force_lowercase=True, device='cpu')
        with gzip.open(states, 'rb') as f:
            states_dict = torch.load(f, map_location=torch.device('cpu'))
        bertnli.load_state_dict(states_dict, strict=False)
        return bertnli

    def predict(self, sent1s, sent2s):
        batches, buf1, buf2 = [], [], []
        for sent1, sent2 in zip(sent1s, sent2s):
            buf1.append(sent1)
            buf2.append(sent2)
            if len(buf1) >= self.batch:
                batches.append((buf1, buf2))
                buf1, buf2 = [], []
        if len(buf1) > 0:
            batches.append((buf1, buf2))

        probs, preds = [], []
        with torch.no_grad():
            for b1, b2 in batches:
                out = self.model(b1, b2)
                out = softmax(out, dim=-1).detach().cpu()
                _, idxs = out.max(dim=-1)
                for i, idx in enumerate(idxs):
                    idx = int(idx)
                    probs.append({'entailment': float(out[i][BERTNLI.LABEL_ENTAILMENT]),
                                  'neutral': float(out[i][BERTNLI.LABEL_NEUTRAL]),
                                  'contradiction': float(out[i][BERTNLI.LABEL_CONTRADICTION])})
                    if idx == BERTNLI.LABEL_ENTAILMENT:
                        preds.append('entailment')
                    elif idx == BERTNLI.LABEL_NEUTRAL:
                        preds.append('neutral')
                    elif idx == BERTNLI.LABEL_CONTRADICTION:
                        preds.append('contradiction')
                    else:
                        raise ValueError('Unknown label index {0}'.format(idx))
        return probs, preds



import os
import json
import torch
import ast
from torch.utils.data import Dataset


def make_lines(root, split, file, processing):
    return [processing(line.strip()) for line in open(os.path.join(root, split + '.' + file))]


class AnyDataset(Dataset):
    def __init__(self,
                 root=None,
                 file=None,
                 split=None,
                 processing=None,
                 name=None,
                 **kwargs):
        assert split is not None, "Argument split cannot be None"

        self.root = root
        self.file = file
        self.split = split
        self.name = name or "any"
        self.processing = eval(processing or 'lambda x: x')
        self.lines = make_lines(root, split, file, self.processing)

    def __getitem__(self, index):
        return {self.name: self.lines[index]}

    def get_collate_fn(self):
        def collate_fn(batch):
            collated = {self.name: [s[self.name] for s in batch]}
            return collated

        return collate_fn

    def __len__(self):
        return len(self.lines)

    def inference(self, sentences):
        raise NotImplementedError()

    def __repr__(self):
        return "AnyDataset\n" + \
            json.dumps({"root": self.root,
                        "file": self.file,
                        "processing": self.processing,
                        "name": self.name
                        }, indent=4,
                       sort_keys=False, default=str)



import json
import os
import torch

from tqdm import tqdm
import numpy as np
from torchvision.utils import save_image
from dalle_pytorch import CLIP as pyCLIP
from torch import nn
import torch.nn.functional as F

from vilmedic.models.utils import get_n_params


def evaluation(models, config, dl, **kwargs):
    activation = {}

    def get_activation(name):
        def hook(model, input, output):
            activation[name] = output.detach()

        return hook

    losses = []
    linguistics = []
    visuals = []
    linguistics_norm = []
    visuals_norm = []

    model = models[0]
    pbar = tqdm(dl, total=len(dl))

    if isinstance(model, torch.nn.DataParallel):
        model.module.clip.to_text_latent.register_forward_hook(get_activation('to_text_latent'))
        model.module.clip.to_visual_latent.register_forward_hook(get_activation('to_visual_latent'))
    else:
        model.clip.to_text_latent.register_forward_hook(get_activation('to_text_latent'))
        model.clip.to_visual_latent.register_forward_hook(get_activation('to_visual_latent'))

    for batch in pbar:
        out = model(**batch)
        out['loss'] = out['loss'].mean()
        losses.append(out['loss'].cpu().data.numpy())

        text_latents = activation['to_text_latent']
        image_latents = activation['to_visual_latent']

        linguistics.append(text_latents.cpu().data.numpy())
        visuals.append(image_latents.cpu().data.numpy())

        text_latents_norm, image_latents_norm = map(lambda t: F.normalize(t, p=2, dim=-1), (text_latents, image_latents))

        linguistics_norm.append(text_latents_norm.cpu().data.numpy())
        visuals_norm.append(image_latents_norm.cpu().data.numpy())

    linguistics = np.concatenate(linguistics, axis=0)
    visuals = np.concatenate(visuals, axis=0)
    linguistics_norm = np.concatenate(linguistics_norm, axis=0)
    visuals_norm = np.concatenate(visuals_norm, axis=0)
    split = dl.dataset.split
    np.save("CLIP_mimic_{}_visual".format(split), visuals)
    np.save("CLIP_mimic_{}_linguistic".format(split), linguistics)
    np.save("CLIP_norm_mimic_{}_visual".format(split), visuals_norm)
    np.save("CLIP_norm_mimic_{}_linguistic".format(split), linguistics_norm)

    return {'loss': np.ndarray.mean(np.array(losses))}


class CLIP(nn.Module):
    def __init__(self,
                 clip,
                 **kwargs):
        super(CLIP, self).__init__()
        self.clip_config = clip
        self.clip = pyCLIP(
            **self.clip_config
        )

        self.eval_func = evaluation

    def forward(self, input_ids, images, **kwargs):
        loss = self.clip(input_ids.cuda(), images.cuda(), text_mask=torch.ones_like(input_ids).bool().cuda(),
                         return_loss=True)
        return {'loss': loss}

    def __repr__(self):
        s = str(type(self.clip).__name__) + '(' + str(
            json.dumps(dict(self.clip_config), indent=4, sort_keys=True)) + ')\n'
        s += "{}\n".format(get_n_params(self))
        return s



import torch
import torch.nn as nn
import torch.nn.functional as F


class VICREGLoss(nn.Module):
    def __init__(self,
                 sim_loss_weight: float = 25.0,
                 var_loss_weight: float = 25.0,
                 cov_loss_weight: float = 1.0,
                 **kwargs):
        super().__init__()
        self.sim_loss_weight = sim_loss_weight
        self.var_loss_weight = var_loss_weight
        self.cov_loss_weight = cov_loss_weight

    def forward(self, z1: torch.Tensor,
                z2: torch.Tensor,
                ) -> torch.Tensor:
        """Computes VICReg's loss given batch of projected features z1 from view 1 and
        projected features z2 from view 2.
        Args:
            z1 (torch.Tensor): NxD Tensor containing projected features from view 1.
            z2 (torch.Tensor): NxD Tensor containing projected features from view 2.
        Returns:
            torch.Tensor: VICReg loss.
        """

        sim_loss = VICREGLoss.invariance_loss(z1, z2)
        var_loss = VICREGLoss.variance_loss(z1, z2)
        cov_loss = VICREGLoss.covariance_loss(z1, z2)

        loss = self.sim_loss_weight * sim_loss + self.var_loss_weight * var_loss + self.cov_loss_weight * cov_loss
        return loss

    @staticmethod
    def invariance_loss(z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Computes mse loss given batch of projected features z1 from view 1 and
        projected features z2 from view 2.
        Args:
            z1 (torch.Tensor): NxD Tensor containing projected features from view 1.
            z2 (torch.Tensor): NxD Tensor containing projected features from view 2.
        Returns:
            torch.Tensor: invariance loss (mean squared error).
        """

        return F.mse_loss(z1, z2)

    @staticmethod
    def variance_loss(z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Computes variance loss given batch of projected features z1 from view 1 and
        projected features z2 from view 2.
        Args:
            z1 (torch.Tensor): NxD Tensor containing projected features from view 1.
            z2 (torch.Tensor): NxD Tensor containing projected features from view 2.
        Returns:
            torch.Tensor: variance regularization loss.
        """

        eps = 1e-4
        std_z1 = torch.sqrt(z1.var(dim=0) + eps)
        std_z2 = torch.sqrt(z2.var(dim=0) + eps)
        std_loss = torch.mean(F.relu(1 - std_z1)) + torch.mean(F.relu(1 - std_z2))
        return std_loss

    @staticmethod
    def covariance_loss(z1: torch.Tensor, z2: torch.Tensor) -> torch.Tensor:
        """Computes covariance loss given batch of projected features z1 from view 1 and
        projected features z2 from view 2.
        Args:
            z1 (torch.Tensor): NxD Tensor containing projected features from view 1.
            z2 (torch.Tensor): NxD Tensor containing projected features from view 2.
        Returns:
            torch.Tensor: covariance regularization loss.
        """

        N, D = z1.size()

        z1 = z1 - z1.mean(dim=0)
        z2 = z2 - z2.mean(dim=0)
        cov_z1 = (z1.T @ z1) / (N - 1)
        cov_z2 = (z2.T @ z2) / (N - 1)

        diag = torch.eye(D, device=z1.device)
        cov_loss = cov_z1[~diag.bool()].pow_(2).sum() / D + cov_z2[~diag.bool()].pow_(2).sum() / D
        return cov_loss

    def __repr__(self):
        return "VICREGLoss(\n" + \
               "\t(sim_loss_weight): {}\n".format(self.sim_loss_weight) + \
               "\t(var_loss_weight): {}\n".format(self.var_loss_weight) + \
               "\t(cov_loss_weight): {}\n".format(self.cov_loss_weight) + \
               ")"



import numpy as np
import torch.nn as nn
import stanza
from stanza import Pipeline

import logging

stanza.download('en', processors='tokenize,lemma,pos,ner')
stanza.download('en', package='radiology')

class RadEntityMatchExact(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()
        self.ner = Pipeline(lang='en', package='radiology', processors={'tokenize': 'default', 'ner': 'radiology'},
                            **{'tokenize_batch_size': 256, 'ner_batch_size': 256})
        self.target_types = {'S-ANATOMY', 'S-OBSERVATION'}

    def forward(self, refs, hyps):
        docs_h = self.ner([stanza.Document([], text=d.lower().replace(' .', '.')) for d in hyps])
        docs_r = self.ner([stanza.Document([], text=d.lower().replace(' .', '.')) for d in refs])

        scores_e = []
        for doc_h, doc_r in zip(docs_h, docs_r):

            # NER
            ner_h = []
            for sentence in doc_h.sentences:
                ner_h.extend([ner['text'] for ner in sentence.to_dict() if ner['ner'] in self.target_types])

            ner_r = []
            for sentence in doc_r.sentences:
                ner_r.extend([ner['text'] for ner in sentence.to_dict() if ner['ner'] in self.target_types])

            # precision
            match_p = sum([1.0 for ner in ner_h if ner in ner_r])
            total_p = len(ner_h)
            pr_e = match_p / total_p if total_p > 0 else 0.0

            # recall
            match_r = sum([1.0 for ner in ner_h if ner in ner_r])
            total_r = len(ner_r)
            rc_e = match_r / total_r if total_r > 0 else 0.0

            # harmonic mean
            score_e = 2 * pr_e * rc_e / (pr_e + rc_e) if pr_e > 0.0 and rc_e > 0.0 else 0.0
            scores_e.append(score_e)

        mean_exact_e = np.mean(scores_e)
        return mean_exact_e, scores_e, docs_h, docs_r


if __name__ == '__main__':
    v = RadEntityMatchExact()
    x = v(hyps=['No pleural effusion. Normal heart size.', 'Normal heart size.'] * 1,
          refs=['No pleural effusions.', 'Enlarged heart.'] * 1)

    print(x[0], x[1])  # 0.7 [0.4, 1.0]



import torch
import torch.nn as nn


def cosine_similarity(x1, x2, dim=1, eps=1e-8):
    """Returns cosine similarity between x1 and x2, computed along dim."""
    w12 = torch.sum(x1 * x2, dim)
    w1 = torch.norm(x1, 2, dim)
    w2 = torch.norm(x2, 2, dim)
    return (w12 / (w1 * w2).clamp(min=eps)).squeeze()


def gloria_attention_fn(query, context, temp1):
    """
    query: batch x ndf x queryL
    context: batch x ndf x ih x iw (sourceL=ihxiw)
    mask: batch_size x sourceL
    """
    batch_size, queryL = query.size(0), query.size(2)
    ih, iw = context.size(2), context.size(3)
    sourceL = ih * iw

    # --> batch x sourceL x ndf
    context = context.view(batch_size, -1, sourceL)
    contextT = torch.transpose(context, 1, 2).contiguous()

    # Get attention
    # (batch x sourceL x ndf)(batch x ndf x queryL)
    # -->batch x sourceL x queryL
    attn = torch.bmm(contextT, query)
    # --> batch*sourceL x queryL
    attn = attn.view(batch_size * sourceL, queryL)
    attn = nn.Softmax(dim=-1)(attn)

    # --> batch x sourceL x queryL
    attn = attn.view(batch_size, sourceL, queryL)
    # --> batch*queryL x sourceL
    attn = torch.transpose(attn, 1, 2).contiguous()
    attn = attn.view(batch_size * queryL, sourceL)

    attn = attn * temp1
    attn = nn.Softmax(dim=-1)(attn)
    attn = attn.view(batch_size, queryL, sourceL)
    # --> batch x sourceL x queryL
    attnT = torch.transpose(attn, 1, 2).contiguous()

    # (batch x ndf x sourceL)(batch x sourceL x queryL)
    # --> batch x ndf x queryL
    weightedContext = torch.bmm(context, attnT)

    return weightedContext, attn.view(batch_size, -1, ih, iw)


def global_loss(cnn_code, rnn_code, eps=1e-8, temp3=10.0):
    batch_size = cnn_code.shape[0]
    labels = torch.arange(batch_size).cuda()

    if cnn_code.dim() == 2:
        cnn_code = cnn_code.unsqueeze(0)
        rnn_code = rnn_code.unsqueeze(0)

    cnn_code_norm = torch.norm(cnn_code, 2, dim=2, keepdim=True)
    rnn_code_norm = torch.norm(rnn_code, 2, dim=2, keepdim=True)

    scores0 = torch.bmm(cnn_code, rnn_code.transpose(1, 2))
    norm0 = torch.bmm(cnn_code_norm, rnn_code_norm.transpose(1, 2))
    scores0 = scores0 / norm0.clamp(min=eps) * temp3

    # --> batch_size x batch_size
    scores0 = scores0.squeeze(0)

    scores1 = scores0.transpose(0, 1)
    loss0 = nn.CrossEntropyLoss()(scores0, labels)
    loss1 = nn.CrossEntropyLoss()(scores1, labels)
    return loss0, loss1


def local_loss(
        img_features, words_emb, cap_lens, temp1=4.0, temp2=5.0, temp3=10.0, agg="sum"
):
    batch_size = img_features.shape[0]

    att_maps = []
    similarities = []
    # cap_lens = cap_lens.data.tolist()
    for i in range(words_emb.shape[0]):

        # Get the i-th text description
        words_num = cap_lens[i]  # 25
        # TODO: remove [SEP]
        # word = words_emb[i, :, 1:words_num+1].unsqueeze(0).contiguous()    # [1, 768, 25]
        word = words_emb[i, :, :words_num].unsqueeze(0).contiguous()  # [1, 768, 25]
        word = word.repeat(batch_size, 1, 1)  # [48, 768, 25]
        context = img_features  # [48, 768, 19, 19]

        weiContext, attn = gloria_attention_fn(
            word, context, temp1
        )  # [48, 768, 25], [48, 25, 19, 19]

        att_maps.append(
            attn[i].unsqueeze(0).contiguous()
        )  # add attention for curr index  [25, 19, 19]
        word = word.transpose(1, 2).contiguous()  # [48, 25, 768]
        weiContext = weiContext.transpose(1, 2).contiguous()  # [48, 25, 768]

        word = word.view(batch_size * words_num, -1)  # [1200, 768]
        weiContext = weiContext.view(batch_size * words_num, -1)  # [1200, 768]

        row_sim = cosine_similarity(word, weiContext)
        row_sim = row_sim.view(batch_size, words_num)  # [48, 25]

        row_sim.mul_(temp2).exp_()
        if agg == "sum":
            row_sim = row_sim.sum(dim=1, keepdim=True)  # [48, 1]
        else:
            row_sim = row_sim.mean(dim=1, keepdim=True)  # [48, 1]
        row_sim = torch.log(row_sim)

        similarities.append(row_sim)

    similarities = torch.cat(similarities, 1)  #
    similarities = similarities * temp3
    similarities1 = similarities.transpose(0, 1)  # [48, 48]

    labels = torch.arange(batch_size).cuda()

    loss0 = nn.CrossEntropyLoss()(similarities, labels)  # labels: arange(batch_size)
    loss1 = nn.CrossEntropyLoss()(similarities1, labels)
    return loss0, loss1, att_maps


class GLoRIALoss(nn.Module):
    def __init__(self, local_loss_weight=1.0, global_loss_weight=1.0, temp1=4.0, temp2=5.0, temp3=10.0):
        super(GLoRIALoss, self).__init__()
        self.local_loss_weight = local_loss_weight
        self.global_loss_weight = global_loss_weight
        self.temp1 = temp1
        self.temp2 = temp2
        self.temp3 = temp3

    def forward(self, global_features, local_features, word_embeddings, sent_embeddings, sents):
        l_loss0, l_loss1, attn_maps = self._calc_local_loss(
            local_features, word_embeddings, sents
        )
        g_loss0, g_loss1 = self._calc_global_loss(global_features, sent_embeddings)

        # weighted loss
        loss = 0
        loss += (l_loss0 + l_loss1) * self.local_loss_weight
        loss += (g_loss0 + g_loss1) * self.global_loss_weight

        return loss, attn_maps

    def _calc_local_loss(self, img_emb_l, text_emb_l, sents):
        cap_lens = [
            len([w for w in sent if not w.startswith("[")]) + 1 for sent in sents
        ]
        l_loss0, l_loss1, attn_maps = local_loss(
            img_emb_l,
            text_emb_l,
            cap_lens,
            temp1=self.temp1,
            temp2=self.temp2,
            temp3=self.temp3,
        )
        return l_loss0, l_loss1, attn_maps

    def _calc_global_loss(self, img_emb_g, text_emb_g):
        g_loss0, g_loss1 = global_loss(img_emb_g, text_emb_g, temp3=self.temp3)
        return g_loss0, g_loss1



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from scipy.special import lambertw


class SuperLoss(nn.Module):
    def __init__(self, C, lam=0.25):
        super(SuperLoss, self).__init__()
        self.tau = torch.log(torch.FloatTensor([C])).cuda()
        self.lam = torch.FloatTensor([lam]).cuda()

    def forward(self, l_i):
        l_i_detach = l_i.detach()
        # self.tau = 0.9 * self.tau + 0.1 * l_i_detach
        sigma = self.sigma(l_i_detach)
        loss = (l_i - self.tau) * sigma + self.lam * torch.log(sigma) ** 2
        loss = loss.mean()
        return loss

    def sigma(self, l_i):
        x = -2 / torch.exp(torch.ones_like(l_i))
        y = 0.5 * torch.max(x, (l_i - self.tau) / self.lam)
        y = y.cpu().numpy()
        sigma = np.exp(-lambertw(y))
        sigma = sigma.real.astype(np.float32)
        sigma = torch.from_numpy(sigma).cuda()
        return sigma


class LabelSmoothingCrossEntropy(nn.Module):
    def __init__(self, smoothing=0.1, reduction='mean', **kwargs):
        super(LabelSmoothingCrossEntropy, self).__init__()
        self.smoothing = smoothing
        self.reduction = reduction

    def forward(self, output, target):
        B, c = output.size()
        log_preds = F.log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_preds.sum()
        else:
            loss = -log_preds.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()
        return loss * self.smoothing / c + (1 - self.smoothing) * F.nll_loss(log_preds, target,
                                                                             reduction=self.reduction)


class LabelSmoothingCrossEntropyWithSuperLoss(nn.Module):
    def __init__(self, classes, eps=0.1, reduction='mean', **kwargs):
        super(LabelSmoothingCrossEntropyWithSuperLoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.classes = classes
        self.super_loss = SuperLoss(C=classes)

    def forward(self, output, target):
        B, c = output.size()
        log_preds = F.log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_preds.sum()
        else:
            loss = -log_preds.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()

        loss_cls = loss * self.eps / c + (1 - self.eps) * self.super_loss(
            F.nll_loss(log_preds, target, reduction='none'))
        return loss_cls

    def __repr__(self):
        s = 'LabelSmoothingCrossEntropyWithSuperLoss (eps=' + str(
            self.eps) + ', reduction=' + self.reduction + ', classes=' + str(self.classes) + ')'
        return s


class MixUpLoss(nn.Module):
    def __init__(self, criterion, **kwargs):
        super().__init__()
        self.criterion = MixUpLoss.get_loss(criterion, **kwargs)

    def forward(self, pred, label, label_mixed, lam):
        label_mixed = label_mixed.cuda()
        lam = lam.cuda()
        return lam * self.criterion(pred, label) + (1 - lam) * self.criterion(pred, label_mixed)

    def __repr__(self):
        s = 'MixUpLoss (criterion=' + str(self.criterion) + ')'
        return s

    @staticmethod
    def get_loss(loss_func, **loss_args):
        if hasattr(nn, loss_func):
            loss_func = getattr(nn, loss_func)(**loss_args)
        else:
            try:
                loss_func = eval(loss_func)(**loss_args)
            except NameError:
                raise NotImplementedError("Loss {} not implemented".format(loss_args))
        return loss_func



import torch
import torch.nn as nn
from vilmedic.models.utils import get_n_params

from vilmedic.blocks.vision import *
from vilmedic.blocks.huggingface.encoder.encoder_model import EncoderModel
from vilmedic.blocks.losses import ConVIRTLoss, InfoNCELoss

from tqdm import tqdm
import numpy as np


def evaluation(models, config, dl, from_training, **kwargs):
    # No ensembling for this evaluation
    model = models[0]

    losses = []
    linguistics = []
    visuals = []

    pbar = tqdm(dl, total=len(dl))
    for i, batch in enumerate(pbar):
        batch = {k: v.cuda() if isinstance(v, torch.Tensor) else v for k, v in batch.items()}
        out = model(**batch)
        losses.append(out['loss'].mean().cpu().data.numpy())

        if not from_training:
            linguistics.append(out['linguistic'].cpu().data)
            visuals.append(out['visual'].cpu().data)

    if from_training:
        return {'loss': np.ndarray.mean(np.array(losses))}

    return {'loss': np.ndarray.mean(np.array(losses)),
            'linguistic': torch.cat(linguistics),
            'visual': torch.cat(visuals)
            }


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class ConVIRT(nn.Module):

    def __init__(self, encoder, cnn, projection, loss, forward_batch_size=256, **kwargs):
        super().__init__()

        # Linguistic encoder
        self.linguistic = EncoderModel(encoder)

        # Visual Encoder
        self.visual = eval(cnn.pop('proto'))(**cnn)

        # Projection
        self.vis_proj = nn.Sequential(
            nn.Linear(projection.visual_embedding_dim, projection.projection_dim),
            nn.ReLU(),
            nn.Linear(projection.projection_dim, projection.projection_dim),
        )
        self.lin_proj = nn.Sequential(
            nn.Linear(projection.textual_embedding_dim, projection.projection_dim),
            nn.ReLU(),
            nn.Linear(projection.projection_dim, projection.projection_dim),
        )

        self.loss_fn = eval(loss.pop("proto"))(**loss)
        self.fbs = forward_batch_size

        # Evaluation
        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, images, **kwargs):
        images = images.cuda()
        input_ids = input_ids.cuda()
        attention_mask = attention_mask.cuda()

        bs = images.shape[0]
        linguistics = []
        visuals = []
        for i in list(chunks(range(bs), min(self.fbs, bs))):
            inp = input_ids[i]
            att = attention_mask[i]
            im = images[i]

            linguistic = self.linguistic(input_ids=inp,
                                         attention_mask=att,
                                         **kwargs)
            linguistic = self.lin_proj(linguistic['pooler_output'])
            visual = self.vis_proj(self.visual(im))

            linguistics.append(linguistic)
            visuals.append(visual)

        linguistics = torch.cat(linguistics)
        visuals = torch.cat(visuals)

        loss, loss_l, loss_v = self.loss_fn(linguistics, visuals)

        return {"loss": loss, "loss_l": loss_l, "loss_v": loss_v, "linguistic": linguistics, "visual": visuals}

    def __repr__(self):
        s = "ConVIRT\n"
        s += str(self.visual) + '\n'
        s += str(self.linguistic) + '\n'
        s += str(self.loss_fn) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s



from vilmedic.models.mvqa.MVQA import MVQA

from vilmedic.models.rrs.SumRNN import SumRNN
from vilmedic.models.rrs.SumHugMulti import SumHugMulti
# from vilmedic.models.rrs.RRS_obsolete import RRS
from vilmedic.models.rrs.RRS import RRS
from vilmedic.models.rrs.RRS_SCST import RRS_SCST
from vilmedic.models.rrs.RRS_SCST import RRS_SCST
from vilmedic.models.rrs.SumHugMono_SCST import SumHugMono_SCST

from vilmedic.models.selfsup.clip.VAE import VAE
from vilmedic.models.selfsup.clip.DALLE import DALLE
from vilmedic.models.selfsup.clip.CLIP import CLIP

from vilmedic.models.rrg.RRG import RRG
from vilmedic.models.rrg.RRG_FORCE import RRG_FORCE
from vilmedic.models.rrg.RRG_MULTI import RRG_MULTI
from vilmedic.models.rrg.RRG_EEG import RRG_EEG
from vilmedic.models.rrg.RRG_PPO import RRG_PPO
from vilmedic.models.rrg.RRG_SCST import RRG_SCST
from vilmedic.models.rrg.RRG_2D_CT import RRG_2D_CT

from vilmedic.models.selfsup.conVIRT import ConVIRT
from vilmedic.models.selfsup.SimCLR import SimCLR
from vilmedic.models.selfsup.GLoRIA import GLoRIA
from vilmedic.models.selfsup.VICREG import VICREG
from vilmedic.models.selfsup.SwaV import SwaV
from vilmedic.models.selfsup.PCL import PCL
from vilmedic.models.selfsup.PCL2 import PCL2




import torch.nn as nn
import torch
from collections import OrderedDict

from vilmedic.blocks.huggingface.encoder_decoder.evaluation import evaluation
from vilmedic.blocks.huggingface.encoder_decoder.encoder_decoder_model import EncoderDecoderModel


class RRS(nn.Module):

    def __init__(self, encoder, decoder, dl, **kwargs):
        super().__init__()
        self.enc_dec = EncoderDecoderModel(encoder, decoder)

        # Evaluation
        self.eval_func = evaluation

        # Tokenizer
        self.tokenizer = dl.dataset.tgt_tokenizer

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask, **kwargs):
        out = self.enc_dec(input_ids=input_ids,
                           attention_mask=attention_mask,
                           decoder_input_ids=decoder_input_ids,
                           decoder_attention_mask=decoder_attention_mask,
                           )

        # print(decoder_input_ids[1])
        # print(decoder_input_ids[1].shape)
        # print(torch.argmax(out["logits"][1], dim=-1).shape)
        # print(self.tokenizer_ss.decode(input_ids[0], skip_special_tokens=True, clean_up_tokenization_spaces=False))
        # hyps = self.enc_dec.enc_dec.generate(input_ids.cuda(),
        #                                      attention_mask=attention_mask.cuda(),
        #                                      max_length=80,
        #                                      )
        # # print(hyps)
        # print("####", self.tokenizer.decode(hyps[1], skip_special_tokens=True, clean_up_tokenization_spaces=False))
        # print("####", self.tokenizer.decode(torch.argmax(out["logits"][1], dim=-1), skip_special_tokens=True, clean_up_tokenization_spaces=False))
        # print("####", self.tokenizer.decode(decoder_input_ids[1], skip_special_tokens=True, clean_up_tokenization_spaces=False))
        # print("####")
        return out

    def __repr__(self):
        return "RRS\n" + str(self.enc_dec)



import os
import pydicom
import numpy as np
import json
import PIL
import skimage
import logging
import torchxrayvision as xrv
from PIL import Image, ImageFile
from torch.utils.data import Dataset
from torchvision.transforms import *
from .utils import load_file
from .papers.open_image import *
from .papers.transforms import *
import nibabel as nib
from pydicom.pixel_data_handlers.util import apply_voi_lut
from torch.utils.data._utils.collate import default_collate as pytorch_default_collate

ImageFile.LOAD_TRUNCATED_IMAGES = True  # Are we sure ?
PIL.Image.MAX_IMAGE_PIXELS = None  # Are we sure ?


def vilmedic_collate(batch, multi_image=None):
    # vilmedic_collate only accepts list of tensors (list of images that are transformed)

    # Return one image
    if not multi_image or multi_image == 1:
        return {'images': torch.stack([s['image'][0] for s in batch])}

    # Return multiple image
    new_batch = []
    new_masks = []
    for sample in batch:
        sample_images = sample['image']
        # Remove image to get to multi_image
        if len(sample_images) > multi_image:
            sample_images = sample_images[:multi_image]
        # Pad with zeros to get to multi_image
        if len(sample_images) < multi_image:
            first_image = sample_images[0]
            for _ in range(multi_image - len(sample_images)):
                sample_images.append(first_image.new_zeros(first_image.size()))
        # Stack
        sample_images = torch.cat([s.unsqueeze(dim=0) for s in sample_images], dim=0)
        sample_mask = (sample_images.sum(dim=(1, 2, 3)) != 0)
        new_batch.append(sample_images)
        new_masks.append(sample_mask)

    collated = {'images': torch.stack(new_batch),
                'images_mask': torch.stack(new_masks)}
    return collated


def default_collate(batch):
    collated = {'images': pytorch_default_collate([s['image'][0] for s in batch]),
                'images_mask': None}
    return collated


def read_images(root, image_path, split, file):
    file_path = os.path.join(root, split + '.' + file)
    if '.npy' in file_path:
        return [[x] for x in np.load(file_path)]

    lines = load_file(file_path)
    return [[os.path.join(image_path, image) for image in line.split(',')] for line in lines]


def get_transforms(split, resize, crop, custom_transform_train, custom_transform_validate, ext, called_by_ensemblor):
    # If called_by_ensemblor, return custom_transform_validate or evaluation transform
    if called_by_ensemblor:
        split = not "train"

    if custom_transform_train is not None and split == "train":
        return eval(custom_transform_train)

    if custom_transform_validate is not None and not split == "train":
        return eval(custom_transform_validate)

    if ext in [".npy", ".npz"]:
        return lambda x: x

    if ext == '.xrv':
        return transforms.Compose([xrv.datasets.XRayCenterCrop(),
                                   xrv.datasets.XRayResizer(crop),
                                   lambda x: torch.from_numpy(x)])

    if split == 'train':
        return transforms.Compose([
            transforms.Resize(resize),
            transforms.RandomCrop(crop),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406),
                                 (0.229, 0.224, 0.225))])
    else:
        return transforms.Compose([
            transforms.Resize((crop, crop)),
            transforms.ToTensor(),
            transforms.Normalize((0.485, 0.456, 0.406),
                                 (0.229, 0.224, 0.225))])


def open_image(image, ext):
    if ext == '.jpg' or ext == '.jpeg':
        return Image.open(image).convert('RGB')

    if ext == '.xrv':
        img = skimage.io.imread(image)
        img = xrv.datasets.normalize(img, 255)
        if len(img.shape) > 2:
            img = img[:, :, 0]
        if len(img.shape) < 2:
            print("error, dimension lower than 2 for image")
        return img[None, :, :]

    if ext == '.nib':
        return nib.load(image).get_fdata().astype('float32')

    if ext == '.png':
        return Image.open(image).convert('RGB')

    if ext == '.dcm':
        ds = pydicom.dcmread(image)
        if 'WindowWidth' in ds:
            img = apply_voi_lut(ds.pixel_array, ds).astype(float)
        else:
            img = ds.pixel_array.astype(float)
        img = (np.maximum(img, 0) / img.max()) * 255.0
        img = np.uint8(img)
        return Image.fromarray(img).convert('RGB')

    if ext in ['.npy', '.npz']:
        if type(image) == str:
            image = np.load(image)
        if type(image) == np.ndarray:
            image = torch.from_numpy(image)
        return image

    # Special cases
    if ext in PAPER_EXT.keys():
        return eval(PAPER_EXT[ext])(image)

    raise NotImplementedError("Image extension {} not implemented".format(ext))


def do_images(images, transform, ext):
    opened_images = [open_image(im, ext) for im in images]
    transformed_images = [transform(im) for im in opened_images]
    return transformed_images


class ImageDataset(Dataset):
    def __init__(self,
                 root=None,
                 file=None,
                 split=None,
                 image_path=None,
                 load_memory=False,
                 custom_transform_train=None,
                 custom_transform_validate=None,
                 resize=256,
                 crop=224,
                 ext='.jpg',
                 multi_image=None,
                 called_by_ensemblor=None,
                 **kwargs):

        assert split is not None, "Argument split cant be None"

        self.root = root
        self.file = file
        self.split = split
        self.image_path = image_path
        self.load_memory = load_memory
        self.resize = eval(str(resize))
        self.crop = eval(str(crop))
        self.ext = ext
        self.multi_image = multi_image or 0
        self.images = None

        if file is not None:
            self.images = read_images(root, image_path, split, file)

        self.transform = get_transforms(split,
                                        self.resize,
                                        self.crop,
                                        custom_transform_train,
                                        custom_transform_validate,
                                        self.ext,
                                        called_by_ensemblor)

    def __len__(self):
        return len(self.images or [])

    def __getitem__(self, index):
        return {'image': do_images(self.images[index], self.transform, self.ext)}

    def inference(self, image):
        if not isinstance(image, (list, str)):
            raise TypeError("Input for image must be str or list")

        if isinstance(image, str):
            image = [[image]]

        if isinstance(image, list):
            for i, im in enumerate(image):
                if isinstance(im, str):
                    image[i] = [im]

        # Image is a list of example, an example a list of images.
        batch = [{'image': do_images(i, self.transform, self.ext)} for i in image]
        return self.get_collate_fn()(batch)

    def get_collate_fn(self):
        def collate_fn(batch):
            try:
                return vilmedic_collate(batch, self.multi_image)
            except TypeError:
                return default_collate(batch)

        return collate_fn

    def __repr__(self):
        transform = self.transform
        if hasattr(self.transform, "transforms"):
            transform = self.transform.transforms

        return "ImageDataset\n" + \
            json.dumps({
                "split": self.split,
                "image_path": self.image_path,
                "root": self.root,
                "file": self.file,
                "transforms": transform,
                "ext": self.ext,
            }, indent=4, sort_keys=False, default=str)



import torch
import torch.nn as nn


class InfoNCELoss(nn.Module):
    def __init__(self, tau, **kwargs):
        super().__init__()
        self.tau = tau
        self.xe = nn.CrossEntropyLoss(reduction='none')

    def forward(self, linguistic, visual):
        n = linguistic.shape[0]
        logits = linguistic @ visual.T
        labels = torch.arange(n).cuda()
        loss_t = self.xe(logits, labels)
        loss_i = self.xe(logits.T, labels)
        loss = (loss_i + loss_t) / 2
        loss = loss.mean()
        return loss, loss_t, loss_i

    def __repr__(self):
        return "InfoNCELoss(\n" + \
               "\t(tau): {}\n".format(self.tau) + \
               ")"



import os
import numpy as np
import json
import torch.nn.functional as F
import torch
import logging
from radgraph import F1RadGraph
from f1chexbert import F1CheXbert
from sklearn.metrics import classification_report, roc_auc_score
from . import *

# RadGraph package overrides logger, need to set back to default
logging.setLoggerClass(logging.Logger)

REWARD_COMPLIANT = {
    "rougel": [RougeL, 1],
    "rouge2": [Rouge2, 1],
    "rouge1": [Rouge1, 1],
    "bleu": [Bleu, 1],
    "meteor": [Meteor, 1],
    "ciderdrl": [CiderDRL, 1],
    "radentitymatchexact": [RadEntityMatchExact, 1],
    "radentitynli": [RadEntityNLI, 1],
    "chexbert": [CheXbert, 1],
    "radgraph": [F1RadGraph, 1],
    "bertscore": [BertScore, 1],
    # "MAUVE": [MauveScorer, 0],
}


def compute_scores(metrics, refs, hyps, split, seed, config, epoch, logger, dump=True):
    scores = dict()
    # If metric is None or empty list
    if metrics is None or not metrics:
        return scores

    assert refs is not None and hyps is not None, \
        "You specified metrics but your evaluation does not return hyps nor refs"

    assert len(refs) == len(hyps), 'refs and hyps must have same length : {} vs {}'.format(len(refs), len(hyps))

    # Dump
    if dump:
        ckpt_dir = config.ckpt_dir
        base = os.path.join(ckpt_dir, '{}_{}_{}'.format(split, seed, '{}'))
        refs_file = base.format('refs.txt')
        hyps_file = base.format('hyps.txt')
        metrics_file = base.format('metrics.txt')

        with open(refs_file, 'w') as f:
            f.write('\n'.join(map(str, refs)))
            f.close()

        with open(hyps_file, 'w') as f:
            f.write('\n'.join(map(str, hyps)))
            f.close()

    for metric in metrics:
        # metric_args = dict()
        #
        # # if metric has arguments
        # if OmegaConf.is_dict(metric):
        #     if len(metric) != 1:
        #         logger.warning("Metric badly formatted: {}. Skipping.".format(metric))
        #         continue
        #     metric_args = metric[list(metric.keys())[0]]
        #     metric = list(metric.keys())[0]

        # Iterating over metrics
        if metric == 'BLEU':
            scores["BLEU"] = Bleu()(refs, hyps)[0]
        elif metric == 'METEOR':
            scores["METEOR"] = Meteor()(refs, hyps)[0]
        elif metric == 'CIDERD':
            scores["CIDERD"] = CiderD()(refs, hyps)[0]
        elif metric == 'bertscore':
            scores["bertscore"] = BertScore()(refs, hyps)[0]
        elif metric in ['ROUGE1', 'ROUGE2', 'ROUGEL']:
            scores[metric] = Rouge(rouges=[metric.lower()])(refs, hyps)[0]
        elif metric == 'MAUVE':
            scores["MAUVE"] = round(
                MauveScorer(config.mauve_featurize_model_name or "distilgpt2")(refs, hyps) * 100, 2)
        elif metric == 'accuracy':
            scores["accuracy"] = round(np.mean(np.array(refs) == np.argmax(hyps, axis=-1)) * 100, 2)
        elif metric == 'f1-score':
            scores["f1-score"] = classification_report(refs, np.argmax(hyps, axis=-1))
        elif metric == 'auroc':
            scores["auroc"] = roc_auc_score(refs, F.softmax(torch.from_numpy(hyps), dim=-1).numpy(), multi_class="ovr")
        elif metric == 'chexbert':
            accuracy, accuracy_per_sample, chexbert_all, chexbert_5 = F1CheXbert(
                refs_filename=base.format('refs.chexbert.txt') if dump else None,
                hyps_filename=base.format('hyps.chexbert.txt') if dump else None) \
                (hyps, refs)
            scores["chexbert-5_micro avg_f1-score"] = chexbert_5["micro avg"]["f1-score"]
            scores["chexbert-all_micro avg_f1-score"] = chexbert_all["micro avg"]["f1-score"]
            scores["chexbert-5_macro avg_f1-score"] = chexbert_5["macro avg"]["f1-score"]
            scores["chexbert-all_macro avg_f1-score"] = chexbert_all["macro avg"]["f1-score"]
        elif metric == 'radentitymatchexact':
            scores["radentitymatchexact"] = RadEntityMatchExact()(refs, hyps)[0]
        elif metric == 'radentitynli':
            scores["radentitynli"] = RadEntityNLI()(refs, hyps)[0]
        elif metric == 'radgraph':
            scores["radgraph_simple"], scores["radgraph_partial"], scores["radgraph_complete"] = \
                F1RadGraph(reward_level="all")(refs=refs, hyps=hyps)[0]
        elif metric == 'stanford_ct_abd_accuracy':
            scores["stanford_ct_abd"] = StanfordCTAbdAcc()(refs=refs, hyps=hyps)[0]
        else:
            logger.warning("Metric not implemented: {}".format(metric))

    if dump:
        with open(metrics_file, 'a+') as f:
            f.write(json.dumps({
                'split': split,
                'epoch': epoch,
                'scores': scores
            }, indent=4, sort_keys=False))
    return scores


if __name__ == '__main__':
    pass



import torch.nn as nn

labels_lists = {
    "radiologist_labels_from_n_grams": {'focal splenic lesions', 'pericholecystic fluid', 'dilated loops of small',
                                        'spleen is enlarged , measuring', 'for biliary duct dilatation',
                                        'gallstones without ct', 'spleen is enlarged',
                                        'gallbladder wall thickening',
                                        'lesion in the pancreatic', 'evidence of cholecystitis', 'surgically absent',
                                        'ct evidence of acute cholecystitis',
                                        'ct evidence of cholecystitis', 'thickening or pericholecystic fluid',
                                        'upper limits of normal',
                                        'prostate is mildly enlarged', 'acute cholecystitis',
                                        'thickening or pericholecystic',
                                        'pancreatic ductal dilatation', 'the prostate is mildly', 'wall thickening',
                                        'findings of cholecystitis',
                                        'bowel obstruction', 'the spleen is enlarged', 'gallstones',
                                        'the upper limits of normal',
                                        'evidence of acute cholecystitis', 'the spleen is enlarged measuring',
                                        'pancreatic ductal dilation',
                                        'gallstones no extrahepatic biliary dilatation',
                                        'wall thickening or pericholecystic',
                                        'calcification is demonstrated the pancreatic', 'of cholecystitis',
                                        'wall thickening or pericholecystic fluid',
                                        'the spleen is enlarged ,', 'lymphadenopathy by size criteria'},
    "chatgpt_labels_from_n_grams": {
        'gallbladder wall',
        'cholecystitis',
        'pericholecystic fluid',
        'ct findings',
        'gallstones',
        'biliary duct dilatation',
        'lymphadenopathy',
        'prostate',
        'uterus',
        'small bowel',
        'bowel obstruction',
        'seminal vesicles',
        'right lower quadrant',
        'spleen',
        'pancreatic duct',
        'pancreatic ductal dilatation',
        'pancreatic head',
        'pancreatic tail',
        'pancreatic body',
        'no focal lesions',
        'no obstruction',
        'normal size',
        'normal appearance',
        'mildly enlarged',
        'enlarged',
        'absent',
        'evidence of',
        'dilated'
    }
}


class StanfordCTAbdAcc(nn.Module):
    def __init__(self, **kwargs):
        super(StanfordCTAbdAcc, self).__init__()

    def forward(self, hyps, refs):

        assert len(refs) == len(hyps), "refs and hyps should have the same length"

        scores = {}

        for labels_name, labels_list in labels_lists.items():
            # Keep track of accuracy per sentence
            accuracy_per_sentence = []
            # Iterate through each pair of sentences
            for ref, hyp in zip(refs, hyps):
                # Find the substrings in the ref sentence
                substrings_in_ref = {label for label in labels_list if label in ref}

                # Count how many of these substrings are also in the hyp sentence
                matching_substrings = sum(1 for label in substrings_in_ref if label in hyp)

                # Calculate accuracy for this sentence
                if substrings_in_ref:
                    sentence_accuracy = matching_substrings / len(substrings_in_ref)

                    # Append to the list of accuracies
                    accuracy_per_sentence.append(sentence_accuracy)

            # Calculate the average accuracy over all sentences
            average_accuracy = sum(accuracy_per_sentence) / len(accuracy_per_sentence) if accuracy_per_sentence else 0

            scores[labels_name] = average_accuracy

        scores["averaged"] = sum([v for v in scores.values()]) / len(list(scores.values()))
        return [scores]


if __name__ == '__main__':
    import random
    import json

    refs = [
        'two dominant rim calcified gallstones , measuring up to 2 . 9 cm , are seen within the lumen of the gallbladder . the gallbladder appears small and contracted , accentuating wall thickening .',
        'prominent mesenteric lymph nodes are noted , measuring up to 1 . 0 cm ( 4 : 192 ).',
        'multiple scattered subcentimeter mesenteric lymph nodes are seen .',
        'punctate calcific densities now seen within the gallbladder ( 3 - 161 ) likely reflecting tiny gallstones , without ct evidence of acute cholecystitis .',
        'distended with layering sludge .', 'layering gallstones without ct evidence of acute cholecystitis .',
        'contracted , with evidence of submucosal edema likely reflecting increased portal venous pressures',
        'distended . no wall thickening or pericholecystic fat stranding .',
        'gallstones . gallbladder is dilated . no findings suggestive of acute cholecystitis .',
        'there are numerous low density lymph nodes within the upper abdomen , primarily surrounding the splenic vein , measuring up to 9 mm . increased number of retroperitoneal nodes , the largest an aortocaval node measures 1 . 0 cm ( 4 : 135 ).',
        'the gallbladder is distended , however , there is no evidence of cholelithiasis or wall thickening to suggest cholecystitis .',
        'contracted . cholecystostomy tube has been removed .',
        'large calcified gallstones at the dependent portion of the gallbladder are noted .',
        'biliary excretion of contrast is visualized in the gallbladder . no gallstones .',
        'air - filled gallstones . otherwise normal appearance of the gallbladder .', 'no radiopaque stones .',
        'mild gallbladder wall thickening which may be related to peritoneal carcinomatosis .',
        'the gallbladder slightly distended measuring up to 4 . 6 cm in maximal diameter , likely related to fasting state . no gallbladder wall thickening , pericholecystic fluid , or radiopaque stones identified .',
        'vicarious excretion of contrast .']

    hyps = random.sample(refs, len(refs))
    scores = {"stanford_ct": StanfordCTAbdAcc()(refs=refs, hyps=hyps)[0]}
    print(json.dumps(scores, indent=4))



import copy
import torch
import torch.nn as nn
from .RRS import RRS

from vilmedic.models.utils import get_n_params
from vilmedic.blocks.rl.SCST import SCST
from vilmedic.blocks.huggingface.encoder_decoder.evaluation import evaluation as evaluation_


def evaluation(models, config, dl, **kwargs):
    models = [m.model if not isinstance(m, nn.DataParallel) else m.module.model for m in
              models]  # Get trained RRS instance
    return evaluation_(models, config, dl, **kwargs)


class RRS_SCST(nn.Module):

    def __init__(self, encoder, decoder, ckpt, dl, scores="ROUGEL", scores_args=None, scores_weights=None, top_k=None,
                 use_nll=False, **kwargs):
        super().__init__()

        # Models
        state_dict = torch.load(ckpt)["model"]
        self.model = RRS(copy.deepcopy(encoder), copy.deepcopy(decoder), dl, **kwargs)
        self.model.load_state_dict(state_dict, strict=True)

        self.encoder = self.model.enc
        self.decoder = self.model.dec.decoder

        # SCST
        self.scst = SCST(decoder=self.decoder,
                         dl=dl,
                         scores=scores,
                         scores_args=scores_args,
                         scores_weights=scores_weights,
                         use_nll=use_nll,
                         top_k=top_k)

        self.eval_func = evaluation

    def forward(self, input_ids, attention_mask, decoder_input_ids, decoder_attention_mask, **kwargs):
        # 1 Greedy
        with torch.no_grad():
            self.model.eval()
            input_ids = input_ids.cuda()
            attention_mask = attention_mask.cuda()
            encoder_outputs = self.encoder(input_ids, attention_mask, return_dict=True)

            reward_greedy, greedy_hyp_list, ref_list = self.scst.forward_greedy(
                input_ids=decoder_input_ids,
                encoder_hidden_states=encoder_outputs.last_hidden_state,
                encoder_attention_mask=attention_mask
            )

        # 2. Sampling
        self.model.train()
        encoder_outputs = self.encoder(input_ids, attention_mask, return_dict=True)
        loss, delta_reward, delta_reward_per_metric, reward_sampling, sampling_hyp_list = self.scst.forward_sampling(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            encoder_hidden_states=encoder_outputs.last_hidden_state,
            encoder_attention_mask=attention_mask,
            reward_greedy=reward_greedy
        )

        return {"loss": loss,
                "custom_print": "reward_sampling {}, "
                                "delta_reward: {},".format(torch.mean(torch.tensor(reward_sampling)),
                                                           torch.mean(torch.tensor(delta_reward))
                                                           ),
                }

    def __repr__(self):
        s = "RRS_SCST\n"
        s += str(self.scst) + '\n'
        s += "{}\n".format(get_n_params(self))
        return s
